"""Shared context for all service classes."""

from __future__ import annotations

from typing import TYPE_CHECKING, Tuple

import grpc

if TYPE_CHECKING:
    from ..config import GqldbConfig
    from ..connection import ConnectionPool
    from ..session import SessionManager
    from ..transaction import TransactionManager


class ServiceContext:
    """Service context holds shared dependencies for all service classes."""

    def __init__(
        self,
        config: GqldbConfig,
        pool: ConnectionPool,
        sessions: SessionManager,
        tx_manager: TransactionManager,
    ):
        """Initialize service context.

        Args:
            config: Client configuration
            pool: Connection pool for gRPC channels
            sessions: Session manager
            tx_manager: Transaction manager
        """
        self.config = config
        self.pool = pool
        self.sessions = sessions
        self.tx_manager = tx_manager

    def get_channel(self) -> grpc.Channel:
        """Get a gRPC channel from the pool."""
        return self.pool.get_connection()

    def get_session_metadata(self) -> Tuple[Tuple[str, str], ...]:
        """Get session metadata for authenticated requests."""
        session = self.sessions.get_session()
        return (('session-id', str(session.id if session else 0)),)

    def update_activity(self) -> None:
        """Update session activity timestamp."""
        self.sessions.update_activity()



"""Type conversion utilities between gRPC proto types and Python types."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List

if TYPE_CHECKING:
    from ..response import Response, Row
    from ..types import (
        GraphInfo,
        GraphType,
        HealthStatus,
        PropertyType,
        TypedValue,
    )


def convert_gql_response(proto_response: Any) -> Response:
    """Convert gRPC response to Response object.

    Args:
        proto_response: Proto GqlResponse

    Returns:
        Response object
    """
    from ..response import Response, Row
    from ..types import PropertyType, TypedValue

    rows = []
    for proto_row in proto_response.rows:
        typed_values = []
        for proto_val in proto_row.values:
            tv = TypedValue(
                type=PropertyType(proto_val.type),
                data=proto_val.data,
                is_null=proto_val.is_null
            )
            typed_values.append(tv)
        rows.append(Row(values=typed_values))

    return Response(
        columns=list(proto_response.columns),
        rows=rows,
        row_count=proto_response.row_count,
        has_more=proto_response.has_more,
        warnings=list(proto_response.warnings),
        rows_affected=proto_response.rows_affected
    )


def convert_graph_info(proto_info: Any) -> GraphInfo:
    """Convert proto GraphInfo to Python GraphInfo.

    Args:
        proto_info: Proto GraphInfo

    Returns:
        GraphInfo object
    """
    from ..types import GraphInfo, GraphType

    return GraphInfo(
        name=proto_info.name,
        graph_type=GraphType(proto_info.graph_type),
        node_count=proto_info.node_count,
        edge_count=proto_info.edge_count,
        description=proto_info.description
    )


def convert_health_status(proto_status: int) -> HealthStatus:
    """Convert proto HealthStatus to enum.

    Args:
        proto_status: Proto health status value

    Returns:
        HealthStatus enum
    """
    from ..types import HealthStatus

    return HealthStatus(proto_status)


def node_data_to_proto(node: Any, proto_module: Any) -> Any:
    """Convert NodeData to proto format.

    Args:
        node: NodeData object
        proto_module: Proto module (gqldb_pb2)

    Returns:
        Proto NodeData
    """
    from ..types import create_typed_value

    properties = {}
    if hasattr(node, 'properties') and node.properties:
        for key, value in node.properties.items():
            properties[key] = create_typed_value(value, proto_module)

    return proto_module.NodeData(
        labels=node.labels if hasattr(node, 'labels') else [],
        properties=properties
    )


def edge_data_to_proto(edge: Any, proto_module: Any) -> Any:
    """Convert EdgeData to proto format.

    Args:
        edge: EdgeData object
        proto_module: Proto module (gqldb_pb2)

    Returns:
        Proto EdgeData
    """
    from ..types import create_typed_value

    properties = {}
    if hasattr(edge, 'properties') and edge.properties:
        for key, value in edge.properties.items():
            properties[key] = create_typed_value(value, proto_module)

    return proto_module.EdgeData(
        label=edge.label if hasattr(edge, 'label') else '',
        from_node_id=edge.from_node_id if hasattr(edge, 'from_node_id') else '',
        to_node_id=edge.to_node_id if hasattr(edge, 'to_node_id') else '',
        properties=properties
    )


def convert_proto_properties(proto_props: Dict[str, Any]) -> Dict[str, Any]:
    """Convert proto properties map to Python dict.

    Args:
        proto_props: Proto properties dict

    Returns:
        Python dict with decoded values
    """
    from ..types import PropertyType, TypedValue

    result = {}
    for key, proto_val in proto_props.items():
        tv = TypedValue(
            type=PropertyType(proto_val.type),
            data=proto_val.data,
            is_null=proto_val.is_null
        )
        result[key] = tv.to_python()
    return result



"""Enumeration types for GQLDB Python driver."""

from enum import IntEnum

class PropertyType(IntEnum):
    """Property type enumeration."""
    UNSET = 0
    INT32 = 1
    UINT32 = 2
    INT64 = 3
    UINT64 = 4
    FLOAT = 5
    DOUBLE = 6
    STRING = 7
    DATETIME = 8  # Deprecated, use TIMESTAMP
    TIMESTAMP = 9
    TEXT = 10
    BLOB = 11
    POINT = 12
    DECIMAL = 13
    LIST = 14
    SET = 15
    MAP = 16
    NULL = 17
    BOOL = 18
    LOCAL_DATETIME = 19
    ZONED_DATETIME = 20
    DATE = 21
    ZONED_TIME = 22
    LOCAL_TIME = 23
    YEAR_TO_MONTH = 24
    DAY_TO_SECOND = 25
    RECORD = 26
    POINT3D = 27
    VECTOR = 28
    TABLE = 29
    PATH = 30
    ERROR = 31
    NODE = 32
    EDGE = 33


class GraphType(IntEnum):
    """Graph type enumeration."""
    OPEN = 0      # Schema-less graph
    CLOSED = 1    # Schema-enforced graph
    ONTOLOGY = 2  # Ontology-enabled graph


class HealthStatus(IntEnum):
    """Health status enumeration."""
    UNKNOWN = 0
    SERVING = 1
    NOT_SERVING = 2
    SERVICE_UNKNOWN = 3


class CacheType(IntEnum):
    """Cache type enumeration."""
    ALL = 0
    AST = 1
    PLAN = 2



"""Regression tests for GQLDB Python SDK.

API contract verification and backward compatibility tests.
Run with: pytest tests/regression/ -v
"""

import os
import pytest

from gqldb.config import ConfigBuilder, GqldbConfig
from gqldb.client import GqldbClient
from gqldb.types import TypedValue
from gqldb.errors import GqldbError


# Test configuration
HOST = os.getenv("GQLDB_HOST", "192.168.1.104:60061")
USERNAME = os.getenv("GQLDB_USERNAME", "admin")
PASSWORD = os.getenv("GQLDB_PASSWORD", "root11")
GRAPH = os.getenv("GQLDB_TEST_GRAPH", "miniCircle")


@pytest.fixture(scope="module")
def client():
    """Create and yield a connected client for the module."""
    config = (
        ConfigBuilder()
        .hosts(HOST)
        .username(USERNAME)
        .password(PASSWORD)
        .default_graph(GRAPH)
        .timeout(30)
        .build()
    )
    c = GqldbClient(config)
    try:
        c.login(USERNAME, PASSWORD)
    except GqldbError as e:
        if "authentication is not enabled" not in str(e).lower():
            raise
    yield c
    c.close()


# ==================== Response Structure Tests ====================


@pytest.mark.regression
class TestResponseStructure:
    """Tests for response data structure contracts."""

    def test_query_response_structure(self, client):
        """Verify query response has expected fields."""
        response = client.gql("RETURN 1 AS value")

        # Verify response is not None
        assert response is not None

        # Verify response has data
        assert not response.is_empty()

        # Verify we can iterate
        count = 0
        for row in response:
            assert row is not None
            count += 1
        assert count > 0

    def test_node_structure(self, client):
        """Verify node data structure contains expected fields."""
        response = client.gql("MATCH (n) RETURN n LIMIT 1")

        if response.is_empty():
            pytest.skip("No nodes in database to test structure")

        nodes, schemas = response.alias("n").as_nodes()
        if len(nodes) == 0:
            pytest.skip("No nodes returned")

        node = nodes[0]

        # Verify node has ID
        assert hasattr(node, 'id') or 'id' in node.__dict__ or hasattr(node, '_id')

        # Verify node has labels
        assert hasattr(node, 'labels') or hasattr(node, '_labels')

        # Verify node has properties
        assert hasattr(node, 'properties') or hasattr(node, '_properties')

    def test_edge_structure(self, client):
        """Verify edge data structure contains expected fields."""
        response = client.gql("MATCH ()-[e]->() RETURN e LIMIT 1")

        if response.is_empty():
            pytest.skip("No edges in database to test structure")

        edges, schemas = response.alias("e").as_edges()
        if len(edges) == 0:
            pytest.skip("No edges returned")

        edge = edges[0]

        # Verify edge has ID
        assert hasattr(edge, 'id') or 'id' in edge.__dict__ or hasattr(edge, '_id')

        # Verify edge has type
        assert hasattr(edge, 'type') or hasattr(edge, '_type') or hasattr(edge, 'label')

    def test_path_structure(self, client):
        """Verify path data structure."""
        response = client.gql("MATCH p = (n)-[e]->(m) RETURN p LIMIT 1")

        if response.is_empty():
            pytest.skip("No paths in database to test structure")

        paths = response.alias("p").as_paths()
        if len(paths) == 0:
            pytest.skip("No paths returned")

        path = paths[0]

        # Verify path has nodes
        assert hasattr(path, 'nodes') and len(path.nodes) > 0


# ==================== API Backward Compatibility Tests ====================


@pytest.mark.regression
class TestAPICompatibility:
    """Tests for API backward compatibility."""

    def test_config_builder_defaults(self):
        """Verify default configuration values."""
        config = ConfigBuilder().hosts(HOST).build()
        assert config is not None

    def test_config_builder_chaining(self):
        """Verify builder pattern chaining."""
        config = (
            ConfigBuilder()
            .hosts(HOST)
            .username(USERNAME)
            .password(PASSWORD)
            .default_graph(GRAPH)
            .timeout(30)
            .build()
        )
        assert config is not None

    def test_client_methods_exist(self, client):
        """Verify client has expected methods."""
        # Core methods
        assert hasattr(client, 'gql')
        assert hasattr(client, 'ping')
        assert hasattr(client, 'health_check')
        assert hasattr(client, 'begin_transaction')
        assert hasattr(client, 'close')

        # Auth methods
        assert hasattr(client, 'login')
        assert hasattr(client, 'logout')


# ==================== Protocol Compatibility Tests ====================


@pytest.mark.regression
class TestProtocolCompatibility:
    """Tests for protocol compatibility."""

    def test_session_id_propagation(self, client):
        """Verify session ID is maintained across requests."""
        # Execute multiple queries
        response1 = client.gql("RETURN 1")
        assert response1 is not None

        response2 = client.gql("RETURN 2")
        assert response2 is not None

    def test_transaction_id_handling(self, client):
        """Verify transaction ID is handled correctly."""
        tx = client.begin_transaction(GRAPH, read_only=True)
        assert tx is not None

        # Execute query within transaction
        from gqldb.client import QueryConfig
        response = client.gql("RETURN 1", config=QueryConfig(transaction_id=tx.id))
        assert response is not None

        # Rollback
        client.rollback(tx.id)


# ==================== Response Method Compatibility Tests ====================


@pytest.mark.regression
class TestResponseMethods:
    """Tests for response method compatibility."""

    def test_response_iteration_methods(self, client):
        """Verify all response iteration methods work."""
        response = client.gql("RETURN 1 AS a, 2 AS b, 3 AS c")

        # Test is_empty
        assert not response.is_empty()

        # Test first
        first = response.first()
        assert first is not None

        # Test last
        last = response.last()
        assert last is not None

        # Test iteration
        count = 0
        for row in response:
            count += 1
        assert count > 0

    def test_response_as_methods_exist(self, client):
        """Verify all As* methods are available via alias()."""
        response = client.gql("RETURN 1 AS value")

        # as_* methods are on AliasedResult, accessed via response.alias() or response.get()
        ar = response.alias("value")
        assert hasattr(ar, 'as_nodes')
        assert hasattr(ar, 'as_edges')
        assert hasattr(ar, 'as_table')
        assert hasattr(ar, 'as_attr')


# ==================== TypedValue Compatibility Tests ====================


@pytest.mark.regression
class TestTypedValueCompatibility:
    """Tests for TypedValue conversion compatibility."""

    @pytest.mark.parametrize("value,name", [
        (True, "bool_true"),
        (False, "bool_false"),
        (42, "int"),
        (3.14, "float"),
        ("hello", "string"),
        (None, "none"),
    ])
    def test_typed_value_conversions(self, value, name):
        """Verify TypedValue conversion methods."""
        tv = TypedValue.from_value(value)
        assert tv is not None

    def test_typed_value_methods_exist(self):
        """Verify TypedValue has expected methods."""
        tv = TypedValue.from_value(42)

        # TypedValue is a dataclass with type, data, is_null properties
        # Use to_python() to convert back to Python value
        assert hasattr(tv, 'type')
        assert hasattr(tv, 'data')
        assert hasattr(tv, 'is_null')
        assert hasattr(tv, 'to_python')

        # Convert back and verify
        result = tv.to_python()
        assert result == 42



"""Performance benchmarks for GQLDB Python SDK.

Run with: pytest tests/benchmark/ -v --benchmark-only
Or manually: pytest tests/benchmark/test_benchmark.py -v
"""

import os
import time
import threading
import statistics
import pytest

from gqldb.config import ConfigBuilder
from gqldb.client import GqldbClient
from gqldb.types import NodeData
from gqldb.errors import GqldbError


# Test configuration
HOST = os.getenv("GQLDB_HOST", "192.168.1.104:60061")
USERNAME = os.getenv("GQLDB_USERNAME", "admin")
PASSWORD = os.getenv("GQLDB_PASSWORD", "root11")
GRAPH = os.getenv("GQLDB_TEST_GRAPH", "miniCircle")


@pytest.fixture(scope="module")
def client():
    """Create a connected client for benchmarks."""
    config = (
        ConfigBuilder()
        .hosts(HOST)
        .username(USERNAME)
        .password(PASSWORD)
        .default_graph(GRAPH)
        .timeout(60)
        .build()
    )
    c = GqldbClient(config)
    try:
        c.login(USERNAME, PASSWORD)
    except GqldbError as e:
        if "authentication is not enabled" not in str(e).lower():
            raise
    yield c
    c.close()


def measure_latency(func, iterations=100):
    """Measure latency statistics for a function."""
    latencies = []
    for _ in range(iterations):
        start = time.perf_counter()
        func()
        latencies.append((time.perf_counter() - start) * 1000)  # ms

    return {
        "min": min(latencies),
        "max": max(latencies),
        "mean": statistics.mean(latencies),
        "p50": statistics.median(latencies),
        "p95": sorted(latencies)[int(len(latencies) * 0.95)],
        "p99": sorted(latencies)[int(len(latencies) * 0.99)],
    }


# ==================== Latency Benchmarks ====================


@pytest.mark.benchmark
class TestLatencyBenchmarks:
    """Latency measurement benchmarks."""

    def test_simple_query_latency(self, client):
        """Measure simple query execution latency."""
        stats = measure_latency(lambda: client.gql("RETURN 1"))

        print(f"\nSimple Query Latency (ms):")
        print(f"  Min: {stats['min']:.2f}")
        print(f"  Max: {stats['max']:.2f}")
        print(f"  Mean: {stats['mean']:.2f}")
        print(f"  P50: {stats['p50']:.2f}")
        print(f"  P95: {stats['p95']:.2f}")
        print(f"  P99: {stats['p99']:.2f}")

        assert stats['mean'] < 1000  # Should be under 1 second

    def test_ping_latency(self, client):
        """Measure ping latency."""
        stats = measure_latency(lambda: client.ping())

        print(f"\nPing Latency (ms):")
        print(f"  Min: {stats['min']:.2f}")
        print(f"  Mean: {stats['mean']:.2f}")
        print(f"  P95: {stats['p95']:.2f}")

    def test_health_check_latency(self, client):
        """Measure health check latency."""
        stats = measure_latency(lambda: client.health_check())

        print(f"\nHealth Check Latency (ms):")
        print(f"  Min: {stats['min']:.2f}")
        print(f"  Mean: {stats['mean']:.2f}")
        print(f"  P95: {stats['p95']:.2f}")


# ==================== Throughput Benchmarks ====================


@pytest.mark.benchmark
class TestThroughputBenchmarks:
    """Throughput measurement benchmarks."""

    def test_query_throughput(self, client):
        """Measure queries per second."""
        iterations = 100
        start = time.perf_counter()

        for _ in range(iterations):
            client.gql("RETURN 1")

        elapsed = time.perf_counter() - start
        qps = iterations / elapsed

        print(f"\nQuery Throughput:")
        print(f"  Queries: {iterations}")
        print(f"  Time: {elapsed:.2f}s")
        print(f"  QPS: {qps:.2f}")

    def test_node_insert_throughput(self, client):
        """Measure node insertion throughput."""
        graph_name = f"bench_insert_{int(time.time() * 1000)}"

        try:
            client.gql(f"CREATE GRAPH {graph_name}")
        except GqldbError:
            pytest.skip("Cannot create test graph")
            return

        try:
            session = client.start_bulk_import(graph_name)

            try:
                iterations = 50
                start = time.perf_counter()

                for i in range(iterations):
                    node = NodeData(labels=["BenchNode"], properties={"id": i})
                    client.insert_nodes(graph_name, [node],
                                       bulk_import_session_id=session.session_id)



                elapsed = time.perf_counter() - start
                nps = iterations / elapsed

                print(f"\nSingle Node Insert Throughput:")
                print(f"  Nodes: {iterations}")
                print(f"  Time: {elapsed:.2f}s")
                print(f"  Nodes/sec: {nps:.2f}")
            finally:
                client.end_bulk_import(session.session_id)

        finally:
            try:
                client.gql(f"DROP GRAPH {graph_name}")
            except GqldbError:
                pass

    def test_batch_insert_throughput(self, client):
        """Measure batch node insertion throughput."""
        graph_name = f"bench_batch_{int(time.time() * 1000)}"
        batch_size = 100
        batches = 10

        try:
            client.gql(f"CREATE GRAPH {graph_name}")
        except GqldbError:
            pytest.skip("Cannot create test graph")
            return

        try:
            session = client.start_bulk_import(graph_name)

            try:
                start = time.perf_counter()

                for i in range(batches):
                    nodes = [
                        NodeData(labels=["BenchNode"], properties={"id": i * batch_size + j})
                        for j in range(batch_size)
                    ]
                    client.insert_nodes(graph_name, nodes,
                                       bulk_import_session_id=session.session_id)



                elapsed = time.perf_counter() - start
                total_nodes = batches * batch_size
                nps = total_nodes / elapsed

                print(f"\nBatch Insert Throughput:")
                print(f"  Batch size: {batch_size}")
                print(f"  Batches: {batches}")
                print(f"  Total nodes: {total_nodes}")
                print(f"  Time: {elapsed:.2f}s")
                print(f"  Nodes/sec: {nps:.2f}")
            finally:
                client.end_bulk_import(session.session_id)

        finally:
            try:
                client.gql(f"DROP GRAPH {graph_name}")
            except GqldbError:
                pass


# ==================== Scalability Benchmarks ====================


@pytest.mark.benchmark
class TestScalabilityBenchmarks:
    """Scalability measurement benchmarks."""

    def test_large_result_set_iteration(self, client):
        """Measure iteration over large result sets."""
        sizes = [100, 1000, 10000]

        for size in sizes:
            try:
                start = time.perf_counter()
                response = client.gql(f"MATCH (n) RETURN n LIMIT {size}")

                count = 0
                for row in response:
                    count += 1

                elapsed = time.perf_counter() - start

                print(f"\nResult Set Iteration (size={size}):")
                print(f"  Rows: {count}")
                print(f"  Time: {elapsed:.3f}s")
                if count > 0:
                    print(f"  Rows/sec: {count / elapsed:.2f}")

            except GqldbError as e:
                print(f"\nResult Set Iteration (size={size}): Skipped - {e}")

    def test_concurrent_queries(self, client):
        """Measure performance with concurrent queries."""
        concurrency_levels = [1, 5, 10]
        queries_per_thread = 20

        for concurrency in concurrency_levels:
            errors = []
            start = time.perf_counter()

            def worker():
                config = (
                    ConfigBuilder()
                    .hosts(HOST)
                    .username(USERNAME)
                    .password(PASSWORD)
                    .default_graph(GRAPH)
                    .timeout(60)
                    .build()
                )
                c = GqldbClient(config)
                try:
                    c.login(USERNAME, PASSWORD)
                except GqldbError as e:
                    if "authentication is not enabled" not in str(e).lower():
                        errors.append(str(e))
                        return

                try:
                    for _ in range(queries_per_thread):
                        c.gql("RETURN 1")
                except GqldbError as e:
                    errors.append(str(e))
                finally:
                    c.close()

            threads = [threading.Thread(target=worker) for _ in range(concurrency)]
            for t in threads:
                t.start()
            for t in threads:
                t.join()

            elapsed = time.perf_counter() - start
            total_queries = concurrency * queries_per_thread
            qps = total_queries / elapsed

            print(f"\nConcurrent Queries (concurrency={concurrency}):")
            print(f"  Total queries: {total_queries}")
            print(f"  Time: {elapsed:.2f}s")
            print(f"  QPS: {qps:.2f}")
            print(f"  Errors: {len(errors)}")


# ==================== Connection Benchmarks ====================


@pytest.mark.benchmark
class TestConnectionBenchmarks:
    """Connection benchmarks."""

    def test_connection_establishment_time(self):
        """Measure connection setup time."""
        iterations = 10
        latencies = []

        for _ in range(iterations):
            start = time.perf_counter()

            config = (
                ConfigBuilder()
                .hosts(HOST)
                .username(USERNAME)
                .password(PASSWORD)
                .default_graph(GRAPH)
                .timeout(30)
                .build()
            )
            c = GqldbClient(config)
            try:
                c.login(USERNAME, PASSWORD)
            except GqldbError:
                pass

            latencies.append((time.perf_counter() - start) * 1000)
            c.close()

        print(f"\nConnection Establishment (ms):")
        print(f"  Min: {min(latencies):.2f}")
        print(f"  Max: {max(latencies):.2f}")
        print(f"  Mean: {statistics.mean(latencies):.2f}")


# ==================== Transaction Benchmarks ====================


@pytest.mark.benchmark
class TestTransactionBenchmarks:
    """Transaction benchmarks."""

    def test_transaction_overhead(self, client):
        """Measure transaction begin/commit overhead."""
        iterations = 50

        # Without transaction
        start = time.perf_counter()
        for _ in range(iterations):
            client.gql("RETURN 1")
        without_tx = time.perf_counter() - start

        # With transaction
        start = time.perf_counter()
        for _ in range(iterations):
            tx = client.begin_transaction(GRAPH, read_only=True)
            from gqldb.client import QueryConfig
            client.gql("RETURN 1", config=QueryConfig(transaction_id=tx.id))
            client.commit(tx.id)
        with_tx = time.perf_counter() - start

        overhead = ((with_tx - without_tx) / without_tx) * 100

        print(f"\nTransaction Overhead:")
        print(f"  Without TX: {without_tx:.2f}s for {iterations} queries")
        print(f"  With TX: {with_tx:.2f}s for {iterations} queries")
        print(f"  Overhead: {overhead:.1f}%")



"""Session Service integration tests."""

import pytest

from gqldb.client import GqldbClient
from gqldb.config import GqldbConfig
from gqldb.errors import GqldbError
from tests.conftest import check_auth_enabled, TEST_USERNAME, TEST_PASSWORD


@pytest.mark.integration
class TestSession:
    """Tests for session operations (login/logout/ping)."""

    def test_login_success(self, client: GqldbClient):
        """Test successful login (skips if auth disabled)."""
        if not check_auth_enabled(client):
            pytest.skip("Authentication is not enabled on this server")

        session = client.login(TEST_USERNAME, TEST_PASSWORD)
        assert session is not None
        assert session.id > 0
        assert client.is_logged_in()
        client.logout()
        client.close()

    def test_login_when_auth_disabled(self, client: GqldbClient):
        """Test login behavior when authentication is disabled."""
        if check_auth_enabled(client):
            pytest.skip("Authentication is enabled on this server")

        from gqldb.errors import GqldbError
        with pytest.raises(GqldbError) as exc_info:
            client.login(TEST_USERNAME, TEST_PASSWORD)
        assert "authentication is not enabled" in str(exc_info.value).lower()
        client.close()

    def test_logout(self, client: GqldbClient):
        """Test logout (skips if auth disabled)."""
        if not check_auth_enabled(client):
            pytest.skip("Authentication is not enabled on this server")

        client.login(TEST_USERNAME, TEST_PASSWORD)
        assert client.is_logged_in()
        client.logout()
        assert not client.is_logged_in()
        client.close()

    def test_client_context_manager(self, test_config: GqldbConfig):
        """Test client as context manager (skips if auth disabled)."""
        with GqldbClient(test_config) as client:
            if not check_auth_enabled(client):
                pytest.skip("Authentication is not enabled on this server")
            client.login(TEST_USERNAME, TEST_PASSWORD)
            assert client.is_logged_in()

    def test_ping(self, connected_client: GqldbClient):
        """Test ping operation."""
        latency = connected_client.ping()
        assert latency >= 0
        print(f"Ping latency: {latency} ns")


@pytest.mark.integration
class TestSessionNoAuth:
    """Tests for server without authentication (192.168.1.104:60062)."""

    def test_ping_without_login(self, no_auth_client: GqldbClient):
        """Test ping works without login on no-auth server."""
        latency = no_auth_client.ping()
        assert latency >= 0
        print(f"No-auth server ping latency: {latency} ns")

    def test_login_returns_error(self, no_auth_client: GqldbClient):
        """Test login returns 'authentication is not enabled' error."""
        with pytest.raises(GqldbError) as exc_info:
            no_auth_client.login(TEST_USERNAME, TEST_PASSWORD)
        assert "authentication is not enabled" in str(exc_info.value).lower()

    def test_query_without_login(self, no_auth_client: GqldbClient):
        """Test query works without login on no-auth server."""
        try:
            resp = no_auth_client.gql("MATCH (n) RETURN n LIMIT 1")
            assert resp is not None
            print(f"Query returned {resp.row_count} rows without login")
        except GqldbError as e:
            # If graph not found, that's ok - the point is we can execute without auth
            if "not found" in str(e).lower():
                print(f"Graph not found (expected on clean server): {e}")
            else:
                raise



"""Session management for GQLDB Python driver."""

from __future__ import annotations

import threading
import time
from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class Session:
    """Represents an authenticated session with a GQLDB server."""
    id: int
    server_version: str
    roles: List[str] = field(default_factory=list)
    default_graph: str = ""
    created_at: float = field(default_factory=time.time)
    last_activity: float = field(default_factory=time.time)
    is_cluster: bool = False
    cluster_id: str = ""
    partition_count: int = 0
    _lock: threading.RLock = field(default_factory=threading.RLock, repr=False)

    def has_role(self, role: str) -> bool:
        """Check if the session has a specific role.

        Args:
            role: Role name to check.

        Returns:
            True if the session has the role.
        """
        with self._lock:
            return role in self.roles

    def idle_duration(self) -> float:
        """Return how long the session has been idle in seconds.

        Returns:
            Idle duration in seconds.
        """
        with self._lock:
            return time.time() - self.last_activity

    def age(self) -> float:
        """Return how long the session has been active in seconds.

        Returns:
            Session age in seconds.
        """
        with self._lock:
            return time.time() - self.created_at


class SessionManager:
    """Manages sessions for the client."""

    def __init__(self):
        self._session: Optional[Session] = None
        self._lock = threading.RLock()

    def login(
        self,
        session_id: int,
        server_version: str,
        roles: List[str],
        default_graph: str,
        is_cluster: bool = False,
        cluster_id: str = "",
        partition_count: int = 0
    ) -> Session:
        """Create a new session.

        Args:
            session_id: Session ID from the server.
            server_version: Server version string.
            roles: List of roles assigned to the user.
            default_graph: Default graph name.
            is_cluster: Whether the server is a cluster.
            cluster_id: Cluster ID if in cluster mode.
            partition_count: Number of partitions in cluster.

        Returns:
            The created Session.
        """
        with self._lock:
            self._session = Session(
                id=session_id,
                server_version=server_version,
                roles=roles,
                default_graph=default_graph,
                is_cluster=is_cluster,
                cluster_id=cluster_id,
                partition_count=partition_count
            )
            return self._session

    def logout(self) -> None:
        """Clear the current session."""
        with self._lock:
            self._session = None

    def get_session(self) -> Optional[Session]:
        """Return the current session.

        Returns:
            Current Session or None.
        """
        with self._lock:
            return self._session

    def get_session_id(self) -> int:
        """Return the current session ID.

        Returns:
            Session ID or 0 if not logged in.
        """
        with self._lock:
            return self._session.id if self._session else 0

    def is_logged_in(self) -> bool:
        """Check if there is an active session.

        Returns:
            True if logged in.
        """
        with self._lock:
            return self._session is not None

    def update_activity(self) -> None:
        """Update the last activity time of the session."""
        with self._lock:
            if self._session:
                self._session.last_activity = time.time()

    def set_default_graph(self, graph: str) -> None:
        """Set the default graph for the session.

        Args:
            graph: Graph name.
        """
        with self._lock:
            if self._session:
                self._session.default_graph = graph

    def get_default_graph(self) -> str:
        """Return the default graph for the session.

        Returns:
            Default graph name or empty string.
        """
        with self._lock:
            return self._session.default_graph if self._session else ""



"""Response handling for GQLDB Python driver."""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, TypeVar

from .types import TypedValue, PropertyType, Schema, PropertyDef, Table, Header, Attr
from .types.graph_models import GqldbNode, GqldbEdge, GqldbPath

T = TypeVar("T")


@dataclass
class Row:
    """Represents a single row in the query result."""
    values: List[TypedValue] = field(default_factory=list)

    def get(self, index: int) -> Any:
        """Get the value at the given column index.

        Args:
            index: Column index.

        Returns:
            The value at the index.

        Raises:
            IndexError: If index is out of range.
        """
        if index < 0 or index >= len(self.values):
            raise IndexError(f"Column index out of range: {index}")
        return self.values[index].to_python()

    def get_string(self, index: int) -> str:
        """Get the value at the given column index as a string.

        Args:
            index: Column index.

        Returns:
            The value as a string.
        """
        val = self.get(index)
        if val is None:
            return ""
        if isinstance(val, str):
            return val
        return str(val)

    def get_int(self, index: int) -> int:
        """Get the value at the given column index as an integer.

        Args:
            index: Column index.

        Returns:
            The value as an integer.
        """
        val = self.get(index)
        if val is None:
            return 0
        if isinstance(val, (int, float)):
            return int(val)
        raise TypeError(f"Cannot convert {type(val)} to int")

    def get_float(self, index: int) -> float:
        """Get the value at the given column index as a float.

        Args:
            index: Column index.

        Returns:
            The value as a float.
        """
        val = self.get(index)
        if val is None:
            return 0.0
        if isinstance(val, (int, float)):
            return float(val)
        raise TypeError(f"Cannot convert {type(val)} to float")

    def get_bool(self, index: int) -> bool:
        """Get the value at the given column index as a boolean.

        Args:
            index: Column index.

        Returns:
            The value as a boolean.
        """
        val = self.get(index)
        if val is None:
            return False
        if isinstance(val, bool):
            return val
        raise TypeError(f"Cannot convert {type(val)} to bool")


@dataclass
class Response:
    """Represents the result of a GQL query."""
    columns: List[str] = field(default_factory=list)
    rows: List[Row] = field(default_factory=list)
    row_count: int = 0
    has_more: bool = False
    warnings: List[str] = field(default_factory=list)
    rows_affected: int = 0

    def is_empty(self) -> bool:
        """Check if the response has no rows.

        Returns:
            True if empty.
        """
        return len(self.rows) == 0

    def first(self) -> Optional[Row]:
        """Get the first row.

        Returns:
            First row or None if empty.
        """
        return self.rows[0] if self.rows else None

    def last(self) -> Optional[Row]:
        """Get the last row.

        Returns:
            Last row or None if empty.
        """
        return self.rows[-1] if self.rows else None

    def get_by_name(self, row: Row, column_name: str) -> Any:
        """Get the value for a column by name.

        Args:
            row: Row to get value from.
            column_name: Column name.

        Returns:
            The value.

        Raises:
            KeyError: If column not found.
        """
        try:
            index = self.columns.index(column_name)
            return row.get(index)
        except ValueError:
            raise KeyError(f"Column not found: {column_name}")

    def __iter__(self) -> Iterator[Row]:
        """Iterate over rows."""
        return iter(self.rows)

    def __len__(self) -> int:
        """Return the number of rows."""
        return len(self.rows)

    def for_each(self, fn: Callable[[Row, int], None]) -> None:
        """Iterate over all rows with a callback.

        Args:
            fn: Callback function receiving (row, index).
        """
        for i, row in enumerate(self.rows):
            fn(row, i)

    def map(self, fn: Callable[[Row], T]) -> List[T]:
        """Transform each row using a callback.

        Args:
            fn: Transformation function.

        Returns:
            List of transformed values.
        """
        return [fn(row) for row in self.rows]

    def to_dicts(self) -> List[Dict[str, Any]]:
        """Convert the response to a list of dictionaries.

        Returns:
            List of dictionaries.
        """
        result = []
        for row in self.rows:
            d = {}
            for i, col in enumerate(self.columns):
                d[col] = row.get(i)
            result.append(d)
        return result

    def to_json(self) -> str:
        """Convert the response to JSON.

        Returns:
            JSON string.
        """
        return json.dumps(self.to_dicts(), default=str)

    def single_value(self) -> Any:
        """Get the single value from a single-row, single-column response.

        Returns:
            The single value.

        Raises:
            ValueError: If response has multiple rows or columns.
        """
        if not self.rows:
            return None
        if len(self.rows) > 1:
            raise ValueError(f"Expected single row, got {len(self.rows)}")
        if not self.rows[0].values:
            return None
        if len(self.rows[0].values) > 1:
            raise ValueError(f"Expected single column, got {len(self.rows[0].values)}")
        return self.rows[0].get(0)

    def single_int(self) -> int:
        """Get the single integer value.

        Returns:
            The single integer value.
        """
        val = self.single_value()
        if val is None:
            return 0
        if isinstance(val, (int, float)):
            return int(val)
        raise TypeError(f"Cannot convert {type(val)} to int")

    def single_string(self) -> str:
        """Get the single string value.

        Returns:
            The single string value.
        """
        val = self.single_value()
        if val is None:
            return ""
        if isinstance(val, str):
            return val
        return str(val)

    def alias(self, column_name: str) -> AliasResult:
        """Get an AliasResult for the specified column name.

        Args:
            column_name: Column name/alias to extract.

        Returns:
            AliasResult for chaining.

        Raises:
            KeyError: If column not found.
        """
        try:
            col_idx = self.columns.index(column_name)
        except ValueError:
            raise KeyError(f"Column alias not found: {column_name}")

        return AliasResult(
            alias=column_name,
            column_idx=col_idx,
            response=self
        )

    def get(self, index: int) -> AliasResult:
        """Get an AliasResult for the column at the specified index.

        Args:
            index: Column index (0-based).

        Returns:
            AliasResult for chaining.

        Raises:
            IndexError: If index out of range.
        """
        if index < 0 or index >= len(self.columns):
            raise IndexError(f"Column index out of range: {index} (total: {len(self.columns)})")

        return AliasResult(
            alias=self.columns[index],
            column_idx=index,
            response=self
        )

    # DEPRECATED: The global as_nodes(), as_edges(), as_paths(), as_table(), as_attr() methods
    # have been removed. Use resp.alias("column").as_nodes() or resp.get(index).as_nodes() instead.

    def _build_schema_from_node(self, label: str, node: "Node") -> Schema:
        """Build schema from a node."""
        properties = []
        for name, value in node.properties.items():
            properties.append(PropertyDef(name=name, type=self._infer_property_type(value)))
        return Schema(name=label, properties=properties)

    def _build_schema_from_edge(self, label: str, edge: "Edge") -> Schema:
        """Build schema from an edge."""
        properties = []
        for name, value in edge.properties.items():
            properties.append(PropertyDef(name=name, type=self._infer_property_type(value)))
        return Schema(name=label, properties=properties)

    def _infer_property_type(self, value: Any) -> PropertyType:
        """Infer property type from value."""
        if value is None:
            return PropertyType.NULL
        if isinstance(value, bool):
            return PropertyType.BOOL
        if isinstance(value, int):
            return PropertyType.INT64
        if isinstance(value, float):
            return PropertyType.DOUBLE
        if isinstance(value, str):
            return PropertyType.STRING
        if isinstance(value, bytes):
            return PropertyType.BLOB
        if isinstance(value, list):
            return PropertyType.LIST
        if isinstance(value, dict):
            return PropertyType.MAP
        return PropertyType.UNSET


@dataclass
class InsertNodesResult:
    """Result of an insert nodes operation."""
    success: bool = False
    node_ids: List[str] = field(default_factory=list)
    node_count: int = 0
    message: str = ""


@dataclass
class InsertEdgesResult:
    """Result of an insert edges operation."""
    success: bool = False
    edge_ids: List[str] = field(default_factory=list)
    edge_count: int = 0
    message: str = ""
    skipped_count: int = 0


@dataclass
class DeleteResult:
    """Result of a delete operation."""
    success: bool = False
    deleted_count: int = 0
    message: str = ""


@dataclass
class ExportConfig:
    """Configuration for the unified Export operation."""
    graph_name: str = ""
    batch_size: int = 0
    export_nodes: bool = True
    export_edges: bool = True
    node_labels: List[str] = field(default_factory=list)
    edge_labels: List[str] = field(default_factory=list)
    include_metadata: bool = False


@dataclass
class ExportStats:
    """Statistics for an export operation."""
    nodes_exported: int = 0
    edges_exported: int = 0
    bytes_written: int = 0
    duration_ms: int = 0


@dataclass
class ExportChunk:
    """A chunk of exported data in JSON Lines format."""
    data: bytes = field(default_factory=bytes)
    is_final: bool = False
    stats: Optional[ExportStats] = None

    def get_data_as_string(self) -> str:
        """Get the data as a UTF-8 string."""
        return self.data.decode('utf-8') if self.data else ""

    def has_stats(self) -> bool:
        """Check if statistics are available."""
        return self.stats is not None


@dataclass
class ExportNodesResult:
    """A batch of exported nodes.

    Deprecated: Use export() with ExportConfig instead.
    """
    nodes: List["ExportedNode"] = field(default_factory=list)
    has_more: bool = False


@dataclass
class ExportEdgesResult:
    """A batch of exported edges.

    Deprecated: Use export() with ExportConfig instead.
    """
    edges: List["ExportedEdge"] = field(default_factory=list)
    has_more: bool = False


@dataclass
class ExportedNode:
    """A node exported from the database.

    Deprecated: Use export() with ExportConfig instead.
    """
    id: str = ""
    labels: List[str] = field(default_factory=list)
    properties: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ExportedEdge:
    """An edge exported from the database.

    Deprecated: Use export() with ExportConfig instead.
    """
    id: str = ""
    label: str = ""
    from_node_id: str = ""
    to_node_id: str = ""
    properties: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Node:
    """Represents a node in the graph database."""
    id: str = ""
    labels: List[str] = field(default_factory=list)
    properties: Dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'schema': self.labels[0] if self.labels else '',
            'uuid': self.properties.get('uuid'),
            'values': dict(self.properties)
        }


@dataclass
class Edge:
    """Represents an edge in the graph database."""
    id: str = ""
    label: str = ""
    from_node_id: str = ""
    to_node_id: str = ""
    properties: Dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'fromId': self.from_node_id,
            'toId': self.to_node_id,
            'schema': self.label,
            'fromUuid': self.properties.get('fromUuid'),
            'toUuid': self.properties.get('toUuid'),
            'uuid': self.properties.get('uuid'),
            'values': dict(self.properties)
        }


@dataclass
class Path:
    """Represents a path in the graph database."""
    nodes: List[Node] = field(default_factory=list)
    edges: List[Edge] = field(default_factory=list)


@dataclass
class AliasResult:
    """Represents a single column's data from the response.

    Allows chaining operations like resp.alias("n2").as_nodes() or resp.get(1).as_nodes().
    """
    alias: str              # Column name
    column_idx: int         # Column index
    response: 'Response'    # Parent response reference

    def _collect_node(self, val: Any, nodes: List[Node], schemas: Dict[str, Schema]) -> None:
        """Convert a GqldbNode to Node and add to the list."""
        if isinstance(val, GqldbNode):
            node = Node(id=val.id, labels=val.labels, properties=val.properties)
            nodes.append(node)
            for label in node.labels:
                if label not in schemas:
                    schemas[label] = self.response._build_schema_from_node(label, node)

    def _collect_edge(self, val: Any, edges: List[Edge], schemas: Dict[str, Schema]) -> None:
        """Convert a GqldbEdge to Edge and add to the list."""
        if isinstance(val, GqldbEdge):
            edge = Edge(id=val.id, label=val.label, from_node_id=val.from_node_id,
                        to_node_id=val.to_node_id, properties=val.properties)
            edges.append(edge)
            if edge.label and edge.label not in schemas:
                schemas[edge.label] = self.response._build_schema_from_edge(edge.label, edge)

    def as_nodes(self) -> Tuple[List[Node], Dict[str, Schema]]:
        """Extract nodes from the aliased column.

        Supports both direct NODE columns and LIST of NODE columns
        (e.g., group variables from quantified patterns).

        Returns:
            Tuple of (list of nodes, schema map).

        Raises:
            TypeError: If the column contains non-null values that are not nodes or lists of nodes.
        """
        nodes: List[Node] = []
        schemas: Dict[str, Schema] = {}

        for row in self.response.rows:
            if self.column_idx >= len(row.values):
                continue

            tv = row.values[self.column_idx]

            if tv.is_null:
                continue

            if tv.type == PropertyType.NODE:
                self._collect_node(tv.to_python(), nodes, schemas)
            elif tv.type == PropertyType.LIST:
                # LIST of NODE (group variable from quantified pattern)
                for item in tv.to_python():
                    self._collect_node(item, nodes, schemas)
            else:
                raise TypeError(
                    f"Type mismatch: column '{self.alias}' (index {self.column_idx}) "
                    f"contains {tv.type}, expected NODE or LIST of NODE"
                )

        return nodes, schemas

    def as_edges(self) -> Tuple[List[Edge], Dict[str, Schema]]:
        """Extract edges from the aliased column.

        Supports both direct EDGE columns and LIST of EDGE columns
        (e.g., group variables from quantified patterns).

        Returns:
            Tuple of (list of edges, schema map).

        Raises:
            TypeError: If the column contains non-null values that are not edges or lists of edges.
        """
        edges: List[Edge] = []
        schemas: Dict[str, Schema] = {}

        for row in self.response.rows:
            if self.column_idx >= len(row.values):
                continue

            tv = row.values[self.column_idx]

            if tv.is_null:
                continue

            if tv.type == PropertyType.EDGE:
                self._collect_edge(tv.to_python(), edges, schemas)
            elif tv.type == PropertyType.LIST:
                # LIST of EDGE (group variable from quantified pattern)
                for item in tv.to_python():
                    self._collect_edge(item, edges, schemas)
            else:
                raise TypeError(
                    f"Type mismatch: column '{self.alias}' (index {self.column_idx}) "
                    f"contains {tv.type}, expected EDGE or LIST of EDGE"
                )

        return edges, schemas

    def as_paths(self) -> List[Path]:
        """Extract paths from the aliased column.

        Returns:
            List of paths.

        Raises:
            TypeError: If the column contains non-null values that are not paths.
        """
        paths: List[Path] = []

        for row in self.response.rows:
            # Boundary check
            if self.column_idx >= len(row.values):
                continue

            tv = row.values[self.column_idx]

            # Skip if null
            if tv.is_null:
                continue

            # Check type mismatch
            if tv.type != PropertyType.PATH:
                raise TypeError(
                    f"Type mismatch: column '{self.alias}' (index {self.column_idx}) "
                    f"contains {tv.type}, expected PATH"
                )

            val = tv.to_python()
            if isinstance(val, GqldbPath):
                # Convert GqldbNode/GqldbEdge to Node/Edge
                nodes = [Node(id=n.id, labels=n.labels, properties=n.properties) for n in val.nodes]
                edges = [Edge(id=e.id, label=e.label, from_node_id=e.from_node_id,
                              to_node_id=e.to_node_id, properties=e.properties) for e in val.edges]
                paths.append(Path(nodes=nodes, edges=edges))

        return paths

    def as_table(self) -> Table:
        """Convert the aliased column to table format.

        If the column contains a GqldbTable value (from the table() function),
        unwraps it into proper Table headers and rows with type inference.

        Returns:
            Table structure.
        """
        from .types.data_types import GqldbTable as GqldbTableType

        # Check if the first non-null value is a GqldbTable — if so, unwrap it
        for row in self.response.rows:
            if self.column_idx < len(row.values):
                tv = row.values[self.column_idx]
                if not tv.is_null and tv.type == PropertyType.TABLE:
                    val = tv.to_python()
                    if isinstance(val, GqldbTableType):
                        return self._unwrap_gqldb_table(val)
            break  # only check first row

        # Fallback: generic single-column table
        headers = [Header(name=self.alias, type=PropertyType.UNSET)]
        rows = []
        for row in self.response.rows:
            if self.column_idx < len(row.values):
                rows.append([row.get(self.column_idx)])
            else:
                rows.append([None])
        return Table(name=self.alias, headers=headers, rows=rows)

    def _unwrap_gqldb_table(self, gtable) -> Table:
        """Unwrap a GqldbTable into a Table with proper headers using original column types."""
        headers = []
        for i, col_name in enumerate(gtable.columns):
            if i < len(gtable.column_types):
                col_type = PropertyType(gtable.column_types[i])
            else:
                col_type = PropertyType.UNSET
            headers.append(Header(name=col_name, type=col_type))

        return Table(name=self.alias, headers=headers, rows=gtable.rows)

    def as_attr(self) -> Attr:
        """Extract scalar or list attribute values from the aliased column.

        Returns:
            Attr with values from the column.
        """
        values = []
        attr_type = PropertyType.UNSET

        for row in self.response.rows:
            if self.column_idx < len(row.values):
                tv = row.values[self.column_idx]
                if attr_type == PropertyType.UNSET and not tv.is_null:
                    attr_type = tv.type
                values.append(tv.to_python())

        return Attr(name=self.alias, type=attr_type, values=values)

    def as_values(self) -> List[Any]:
        """Extract raw Python values from the aliased column.

        Returns:
            List of Python values.
        """
        values = []

        for row in self.response.rows:
            if self.column_idx < len(row.values):
                values.append(row.get(self.column_idx))

        return values



"""Barrel export for all types in GQLDB Python driver."""

# Enums
from .enums import PropertyType, GraphType, HealthStatus, CacheType

# Wrappers
from .wrappers import Int32, UInt32, Float32, UInt64

# Data types
from .data_types import (
    Point,
    Point3D,
    GqldbDecimal,
    GqldbLocalDateTime,
    GqldbDate,
    GqldbZonedDateTime,
    GqldbLocalTime,
    GqldbZonedTime,
    YearToMonth,
    DayToSecond,
    Vector,
    GqldbTable,
    ComputeTopologyResult,
)

# Graph models
from .graph_models import (
    GqldbNode,
    GqldbEdge,
    GqldbPath,
    GqldbError,
    NodeData,
    EdgeData,
    ExportedNode,
    ExportedEdge,
)

# Metadata
from .metadata import (
    GraphInfo,
    TransactionInfo,
    ASTCacheStats,
    PlanCacheStats,
    CacheStats,
    Statistics,
)

# Bulk import
from .bulk_import import (
    BulkCreateNodesOptions,
    BulkCreateEdgesOptions,
    CompactResult,
    BulkImportSession,
    CheckpointResult,
    EndBulkImportResult,
    AbortBulkImportResult,
    BulkImportStatus,
)

# Schema
from .schema import (
    PropertyDef,
    Schema,
    Header,
    Table,
    Attr,
)

# TypedValue and Parameter
from .typed_value import TypedValue, Parameter

__all__ = [
    # Enums
    'PropertyType',
    'GraphType',
    'HealthStatus',
    'CacheType',
    # Wrappers
    'Int32',
    'UInt32',
    'Float32',
    'UInt64',
    # Data types
    'Parameter',
    'Point',
    'Point3D',
    'GqldbDecimal',
    'GqldbLocalDateTime',
    'GqldbDate',
    'GqldbZonedDateTime',
    'GqldbLocalTime',
    'GqldbZonedTime',
    'YearToMonth',
    'DayToSecond',
    'Vector',
    'GqldbTable',
    'ComputeTopologyResult',
    # Graph models
    'GqldbNode',
    'GqldbEdge',
    'GqldbPath',
    'GqldbError',
    'NodeData',
    'EdgeData',
    'ExportedNode',
    'ExportedEdge',
    # Metadata
    'GraphInfo',
    'TransactionInfo',
    'ASTCacheStats',
    'PlanCacheStats',
    'CacheStats',
    'Statistics',
    # Bulk import
    'BulkCreateNodesOptions',
    'BulkCreateEdgesOptions',
    'CompactResult',
    'BulkImportSession',
    'CheckpointResult',
    'EndBulkImportResult',
    'AbortBulkImportResult',
    'BulkImportStatus',
    # Schema
    'PropertyDef',
    'Schema',
    'Header',
    'Table',
    'Attr',
    # TypedValue
    'TypedValue',
]



"""Error definitions for GQLDB Python driver."""


class GqldbError(Exception):
    """Base exception for GQLDB errors."""

    def __init__(self, message: str, code: int = 0, cause: Exception | None = None):
        super().__init__(message)
        self.code = code
        self.cause = cause


# Configuration errors
class NoHostsError(GqldbError):
    """No hosts configured."""

    def __init__(self):
        super().__init__("No hosts configured")


class InvalidTimeoutError(GqldbError):
    """Invalid timeout value."""

    def __init__(self):
        super().__init__("Invalid timeout value")


# Connection errors
class NoConnectionError(GqldbError):
    """No connection available."""

    def __init__(self):
        super().__init__("No connection available")


class ConnectionClosedError(GqldbError):
    """Connection closed."""

    def __init__(self):
        super().__init__("Connection closed")


class ConnectionFailedError(GqldbError):
    """Connection failed."""

    def __init__(self, message: str = "Connection failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


class AllHostsFailedError(GqldbError):
    """All hosts failed to connect."""

    def __init__(self):
        super().__init__("All hosts failed to connect")


class HealthCheckFailedError(GqldbError):
    """Health check failed."""

    def __init__(self):
        super().__init__("Health check failed")


# Session errors
class NotLoggedInError(GqldbError):
    """Not logged in."""

    def __init__(self):
        super().__init__("Not logged in")


class LoginFailedError(GqldbError):
    """Login failed."""

    def __init__(self, message: str = "Login failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


class LogoutFailedError(GqldbError):
    """Logout failed."""

    def __init__(self, message: str = "Logout failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


class SessionExpiredError(GqldbError):
    """Session expired."""

    def __init__(self):
        super().__init__("Session expired")


class InvalidSessionError(GqldbError):
    """Invalid session."""

    def __init__(self):
        super().__init__("Invalid session")


# Transaction errors
class NoTransactionError(GqldbError):
    """No active transaction."""

    def __init__(self):
        super().__init__("No active transaction")


class TransactionFailedError(GqldbError):
    """Transaction failed."""

    def __init__(self, message: str = "Transaction failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


class TransactionNotFoundError(GqldbError):
    """Transaction not found."""

    def __init__(self):
        super().__init__("Transaction not found")


class TransactionAlreadyOpenError(GqldbError):
    """Transaction already open."""

    def __init__(self):
        super().__init__("Transaction already open")


# Query errors
class QueryFailedError(GqldbError):
    """Query failed."""

    def __init__(self, message: str = "Query failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


class QueryTimeoutError(GqldbError):
    """Query timeout."""

    def __init__(self):
        super().__init__("Query timeout")


class InvalidQueryError(GqldbError):
    """Invalid query."""

    def __init__(self, message: str = "Invalid query"):
        super().__init__(message)


class EmptyQueryError(GqldbError):
    """Empty query."""

    def __init__(self):
        super().__init__("Empty query")


# Graph errors
class GraphNotFoundError(GqldbError):
    """Graph not found."""

    def __init__(self, graph_name: str = ""):
        message = f"Graph not found: {graph_name}" if graph_name else "Graph not found"
        super().__init__(message)


class GraphExistsError(GqldbError):
    """Graph already exists."""

    def __init__(self, graph_name: str = ""):
        message = f"Graph already exists: {graph_name}" if graph_name else "Graph already exists"
        super().__init__(message)


class CreateGraphFailedError(GqldbError):
    """Create graph failed."""

    def __init__(self, message: str = "Create graph failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


class DropGraphFailedError(GqldbError):
    """Drop graph failed."""

    def __init__(self, message: str = "Drop graph failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


# Data errors
class InsertFailedError(GqldbError):
    """Insert failed."""

    def __init__(self, message: str = "Insert failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


class DeleteFailedError(GqldbError):
    """Delete failed."""

    def __init__(self, message: str = "Delete failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


class ExportFailedError(GqldbError):
    """Export failed."""

    def __init__(self, message: str = "Export failed", cause: Exception | None = None):
        super().__init__(message, cause=cause)


# Type errors
class InvalidTypeError(GqldbError):
    """Invalid type."""

    def __init__(self, message: str = "Invalid type"):
        super().__init__(message)


class TypeConversionError(GqldbError):
    """Type conversion failed."""

    def __init__(self, message: str = "Type conversion failed"):
        super().__init__(message)


class UnsupportedTypeError(GqldbError):
    """Unsupported type."""

    def __init__(self, type_name: str = ""):
        message = f"Unsupported type: {type_name}" if type_name else "Unsupported type"
        super().__init__(message)



"""Fault-directed tests for GQLDB Python SDK.

Tests error handling and failure scenarios.
Run with: pytest tests/fault/ -v
"""

import os
import time
import pytest

from gqldb.config import ConfigBuilder, GqldbConfig
from gqldb.client import GqldbClient
from gqldb.errors import GqldbError


# Test configuration
AUTH_HOST = os.getenv("GQLDB_HOST", "192.168.1.104:60061")
NO_AUTH_HOST = os.getenv("GQLDB_NO_AUTH_HOST", "192.168.1.104:60062")
USERNAME = os.getenv("GQLDB_USERNAME", "admin")
PASSWORD = os.getenv("GQLDB_PASSWORD", "root11")
GRAPH = os.getenv("GQLDB_TEST_GRAPH", "miniCircle")


# ==================== Network Failure Tests ====================


@pytest.mark.fault
class TestNetworkFailures:
    """Tests for network failure scenarios."""

    def test_connection_timeout(self):
        """Test that client handles connection timeout gracefully."""
        config = (
            ConfigBuilder()
            .hosts("192.0.2.1:60061")  # Non-routable IP (RFC 5737)
            .username(USERNAME)
            .password(PASSWORD)
            .default_graph(GRAPH)
            .timeout(2)  # Very short timeout
            .build()
        )

        try:
            client = GqldbClient(config)
            client.ping()
            pytest.fail("Expected timeout error")
        except Exception as e:
            # Expected - connection should fail
            assert e is not None

    def test_connection_refused(self):
        """Test proper error when server is unreachable."""
        config = (
            ConfigBuilder()
            .hosts("localhost:59999")  # Unlikely to have a service
            .username(USERNAME)
            .password(PASSWORD)
            .default_graph(GRAPH)
            .timeout(5)
            .build()
        )

        try:
            client = GqldbClient(config)
            client.ping()
            pytest.fail("Expected connection refused error")
        except Exception as e:
            # Expected
            assert e is not None

    def test_dns_failure(self):
        """Test invalid hostname error handling."""
        config = (
            ConfigBuilder()
            .hosts("invalid.hostname.that.does.not.exist.local:60061")
            .username(USERNAME)
            .password(PASSWORD)
            .default_graph(GRAPH)
            .timeout(5)
            .build()
        )

        try:
            client = GqldbClient(config)
            client.ping()
            pytest.fail("Expected DNS resolution error")
        except Exception as e:
            # Expected
            assert e is not None


# ==================== Authentication Failure Tests ====================


@pytest.mark.fault
class TestAuthenticationFailures:
    """Tests for authentication failure scenarios."""

    def test_invalid_credentials(self):
        """Test wrong username/password handling."""
        config = (
            ConfigBuilder()
            .hosts(AUTH_HOST)
            .username("invalid_user")
            .password("wrong_password")
            .default_graph(GRAPH)
            .timeout(30)
            .build()
        )

        client = GqldbClient(config)
        try:
            client.login("invalid_user", "wrong_password")
            # If we get here, either auth is disabled or credentials worked
        except GqldbError as e:
            if "authentication is not enabled" in str(e).lower():
                pytest.skip("Authentication is not enabled on this server")
            # Expected authentication error
            assert "invalid" in str(e).lower() or "unauthorized" in str(e).lower() or True
        finally:
            client.close()

    def test_empty_credentials(self):
        """Test login with empty credentials."""
        config = (
            ConfigBuilder()
            .hosts(AUTH_HOST)
            .default_graph(GRAPH)
            .timeout(30)
            .build()
        )

        client = GqldbClient(config)
        try:
            client.login("", "")
            # May succeed if auth is disabled
        except GqldbError as e:
            # Expected error with empty credentials
            assert e is not None
        finally:
            client.close()


# ==================== Transaction Failure Tests ====================


@pytest.mark.fault
class TestTransactionFailures:
    """Tests for transaction failure scenarios."""

    @pytest.fixture
    def client(self):
        """Create and yield a connected client."""
        config = (
            ConfigBuilder()
            .hosts(AUTH_HOST)
            .username(USERNAME)
            .password(PASSWORD)
            .default_graph(GRAPH)
            .timeout(30)
            .build()
        )
        c = GqldbClient(config)
        try:
            c.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise
        yield c
        c.close()

    def test_transaction_timeout(self, client):
        """Test transaction timeout behavior."""
        tx = client.begin_transaction(GRAPH, read_only=False)
        assert tx is not None

        # Clean up
        client.rollback(tx.id)

    def test_rollback_after_error(self, client):
        """Test auto-rollback behavior after query errors."""
        tx = client.begin_transaction(GRAPH, read_only=False)

        # Execute an invalid query to trigger error
        try:
            from gqldb.client import QueryConfig
            client.gql("INVALID SYNTAX QUERY!!!", config=QueryConfig(transaction_id=tx.id))
        except GqldbError:
            pass  # Expected

        # Rollback should still work
        try:
            client.rollback(tx.id)
        except GqldbError as e:
            # May fail if already rolled back
            pass

    def test_commit_on_closed_transaction(self, client):
        """Test commit after rollback error handling."""
        tx = client.begin_transaction(GRAPH, read_only=False)

        # Rollback first
        client.rollback(tx.id)

        # Try to commit after rollback - should fail
        try:
            client.commit(tx.id)
            # May succeed depending on implementation
        except GqldbError:
            # Expected
            pass


# ==================== Bulk Import Failure Tests ====================


@pytest.mark.fault
class TestBulkImportFailures:
    """Tests for bulk import failure scenarios."""

    @pytest.fixture
    def client(self):
        """Create and yield a connected client."""
        config = (
            ConfigBuilder()
            .hosts(AUTH_HOST)
            .username(USERNAME)
            .password(PASSWORD)
            .default_graph(GRAPH)
            .timeout(60)
            .build()
        )
        c = GqldbClient(config)
        try:
            c.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise
        yield c
        c.close()

    def test_bulk_import_abort(self, client):
        """Test abort mid-import and verify cleanup."""
        graph_name = f"abort_test_{int(time.time())}"

        try:
            # Create test graph
            client.gql(f"CREATE GRAPH {graph_name}")
        except GqldbError:
            pytest.skip("Could not create test graph")
            return

        try:
            # Start bulk import
            session = client.start_bulk_import(graph_name)
            if session is None:
                pytest.skip("Bulk import not supported")
                return

            # Abort
            client.abort_bulk_import(session.session_id)
        finally:
            # Cleanup
            try:
                client.gql(f"DROP GRAPH {graph_name}")
            except GqldbError:
                pass

    def test_bulk_import_invalid_data(self, client):
        """Test handling of malformed data in bulk import."""
        # This test verifies error handling for invalid data
        graph_name = f"invalid_data_test_{int(time.time())}"

        try:
            client.gql(f"CREATE GRAPH {graph_name}")
        except GqldbError:
            pytest.skip("Could not create test graph")
            return

        try:
            session = client.start_bulk_import(graph_name)
            if session is None:
                pytest.skip("Bulk import not supported")
                return

            # Try to insert invalid data (missing required fields)
            try:
                from gqldb.types import NodeData
                # Insert node with potentially invalid structure
                nodes = [NodeData(labels=["InvalidNode"], properties={})]
                client.insert_nodes(graph_name, nodes,
                                   bulk_import_session_id=session.session_id)
            except GqldbError:
                # Expected for invalid data
                pass

            # Abort/end session
            try:
                client.abort_bulk_import(session.session_id)
            except GqldbError:
                pass
        finally:
            try:
                client.gql(f"DROP GRAPH {graph_name}")
            except GqldbError:
                pass


# ==================== Query Failure Tests ====================


@pytest.mark.fault
class TestQueryFailures:
    """Tests for query failure scenarios."""

    @pytest.fixture
    def client(self):
        """Create and yield a connected client."""
        config = (
            ConfigBuilder()
            .hosts(AUTH_HOST)
            .username(USERNAME)
            .password(PASSWORD)
            .default_graph(GRAPH)
            .timeout(30)
            .build()
        )
        c = GqldbClient(config)
        try:
            c.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise
        yield c
        c.close()

    def test_invalid_query_syntax(self, client):
        """Test handling of invalid query syntax."""
        with pytest.raises(GqldbError):
            client.gql("THIS IS NOT VALID GQL SYNTAX !!!")

    def test_query_non_existent_graph(self):
        """Test query against non-existent graph."""
        config = (
            ConfigBuilder()
            .hosts(AUTH_HOST)
            .username(USERNAME)
            .password(PASSWORD)
            .default_graph("non_existent_graph_12345")
            .timeout(30)
            .build()
        )

        client = GqldbClient(config)
        try:
            client.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        try:
            client.gql("MATCH (n) RETURN n LIMIT 1")
            # May succeed if default graph is used
        except GqldbError:
            # Expected for non-existent graph
            pass
        finally:
            client.close()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="ultipa",
    version="6.0.2",
    description="Pure Python Ultipa Driver",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ultipa Team",
    url="https://github.com/ultipa/gqldb-drivers",
    packages=find_packages(),
    install_requires=[
        "grpcio>=1.60.0",
        "protobuf>=4.25.0",
    ],
    extras_require={
        "dev": [
            "grpcio-tools>=1.60.0",
            "pytest>=7.4.0",
            "pytest-cov>=4.1.0",
            "pytest-asyncio>=0.23.0",
        ]
    },
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



"""TypedValue encoding and decoding for GQLDB Python driver."""

from __future__ import annotations

import calendar
import json
import math
import re
import struct
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta, timezone
from typing import Any, Dict, List, Optional, Set, Union

from .enums import PropertyType
from .wrappers import Int32, UInt32, Float32, UInt64
from .data_types import (
    Point,
    Point3D,
    GqldbDecimal,
    GqldbLocalDateTime,
    GqldbDate,
    GqldbZonedDateTime,
    GqldbLocalTime,
    GqldbZonedTime,
    YearToMonth,
    DayToSecond,
    Vector,
    GqldbTable,
)
from .graph_models import GqldbNode, GqldbEdge, GqldbPath, GqldbError

@dataclass
class TypedValue:
    """Represents a typed value for wire transmission."""
    type: PropertyType
    data: bytes
    is_null: bool = False

    @classmethod
    def from_value(cls, value: Any) -> "TypedValue":
        """Create a TypedValue from a Python value."""
        if isinstance(value, TypedValue):
            return value

        if value is None:
            return cls(type=PropertyType.NULL, data=b"", is_null=True)

        if isinstance(value, bool):
            data = struct.pack("<B", 1 if value else 0)
            return cls(type=PropertyType.BOOL, data=data)

        # Handle explicit type wrappers (must check before generic int/float)
        if isinstance(value, Int32):
            data = struct.pack("<i", value.value)
            return cls(type=PropertyType.INT32, data=data)

        if isinstance(value, UInt32):
            data = struct.pack("<I", value.value)
            return cls(type=PropertyType.UINT32, data=data)

        if isinstance(value, Float32):
            data = struct.pack("<f", value.value)
            return cls(type=PropertyType.FLOAT, data=data)

        if isinstance(value, UInt64):
            data = struct.pack("<Q", value.value)
            return cls(type=PropertyType.UINT64, data=data)

        if isinstance(value, int):
            # Use int64 for integers
            data = struct.pack("<q", value)
            return cls(type=PropertyType.INT64, data=data)

        if isinstance(value, float):
            data = struct.pack("<d", value)
            return cls(type=PropertyType.DOUBLE, data=data)

        if isinstance(value, str):
            return cls(type=PropertyType.STRING, data=value.encode("utf-8"))

        if isinstance(value, bytes):
            return cls(type=PropertyType.BLOB, data=value)

        if isinstance(value, datetime):
            if value.tzinfo is not None:
                # datetime with tzinfo -> ZONED_DATETIME
                offset_td = value.utcoffset()
                offset_minutes = int(offset_td.total_seconds()) // 60
                nanos = value.microsecond * 1000
                data = struct.pack("<HBBBBBIh", value.year, value.month, value.day,
                                   value.hour, value.minute, value.second,
                                   nanos, offset_minutes)
                return cls(type=PropertyType.ZONED_DATETIME, data=data)
            else:
                # datetime without tzinfo -> LOCAL_DATETIME
                nanos = value.microsecond * 1000
                data = struct.pack("<HBBBBBI", value.year, value.month, value.day,
                                   value.hour, value.minute, value.second, nanos)
                return cls(type=PropertyType.LOCAL_DATETIME, data=data)

        if isinstance(value, date):
            # Binary format: year(2), month(1), day(1), pad(4)
            data = struct.pack("<HBB4x", value.year, value.month, value.day)
            return cls(type=PropertyType.DATE, data=data)

        if isinstance(value, time):
            if value.tzinfo is not None:
                offset_td = value.utcoffset()
                offset_minutes = int(offset_td.total_seconds()) // 60
                nanos = value.microsecond * 1000
                data = struct.pack("<BBBxIh", value.hour, value.minute, value.second,
                                   nanos, offset_minutes)
                return cls(type=PropertyType.ZONED_TIME, data=data)
            else:
                nanos = value.microsecond * 1000
                data = struct.pack("<BBBxI", value.hour, value.minute, value.second, nanos)
                return cls(type=PropertyType.LOCAL_TIME, data=data)

        if isinstance(value, timedelta):
            total_secs = int(value.total_seconds())
            remaining_micros = value.microseconds  # microsecond component only
            nanos = remaining_micros * 1000
            # timedelta stores as days + seconds + microseconds internally
            total_whole_secs = value.days * 86400 + value.seconds
            data = struct.pack("<QI", total_whole_secs, nanos)
            return cls(type=PropertyType.DAY_TO_SECOND, data=data)

        if isinstance(value, list):
            return cls._encode_list(value)

        if isinstance(value, set):
            return cls._encode_set(value)

        if isinstance(value, Point):
            # Binary format: longitude, latitude (8 bytes each)
            data = struct.pack("<dd", value.longitude, value.latitude)
            return cls(type=PropertyType.POINT, data=data)

        if isinstance(value, GqldbDate):
            # Binary format: year(2), month(1), day(1), pad(4)
            data = struct.pack("<HBB4x", value.year, value.month, value.day)
            return cls(type=PropertyType.DATE, data=data)

        if isinstance(value, Point3D):
            data = struct.pack("<ddd", value.x, value.y, value.z)
            return cls(type=PropertyType.POINT3D, data=data)

        if isinstance(value, GqldbDecimal):
            return cls(type=PropertyType.DECIMAL, data=value.value.encode("utf-8"))

        if isinstance(value, GqldbLocalDateTime):
            # Component-based encoding: 11 bytes
            # [year:2 LE][month:1][day:1][hour:1][min:1][sec:1][nanos:4 LE]
            data = struct.pack("<HBBBBBI", value.year, value.month, value.day,
                               value.hour, value.minute, value.second, value.nanosecond)
            return cls(type=PropertyType.LOCAL_DATETIME, data=data)

        if isinstance(value, GqldbZonedDateTime):
            # Component-based encoding: 13 bytes
            # [year:2 LE][month:1][day:1][hour:1][min:1][sec:1][nanos:4 LE][offset:2 LE]
            data = struct.pack("<HBBBBBIh", value.year, value.month, value.day,
                               value.hour, value.minute, value.second,
                               value.nanosecond, value.offset_minutes)
            return cls(type=PropertyType.ZONED_DATETIME, data=data)

        if isinstance(value, GqldbLocalTime):
            # Binary format: hour(1), minute(1), second(1), pad(1), nanosecond(4)
            data = struct.pack("<BBBxI", value.hour, value.minute, value.second, value.nanosecond)
            return cls(type=PropertyType.LOCAL_TIME, data=data)

        if isinstance(value, GqldbZonedTime):
            # Binary format: hour(1), minute(1), second(1), pad(1), nanosecond(4), offset_minutes(2)
            data = struct.pack("<BBBxIh", value.hour, value.minute, value.second, value.nanosecond, value.offset_minutes)
            return cls(type=PropertyType.ZONED_TIME, data=data)

        if isinstance(value, YearToMonth):
            data = struct.pack("<i", value.months)
            return cls(type=PropertyType.YEAR_TO_MONTH, data=data)

        if isinstance(value, DayToSecond):
            # Binary format: seconds(8), nanoseconds(4)
            data = struct.pack("<QI", value.seconds, value.nanoseconds)
            return cls(type=PropertyType.DAY_TO_SECOND, data=data)

        if isinstance(value, Vector):
            data = bytearray()
            data.extend(struct.pack("<I", len(value.values)))
            for v in value.values:
                data.extend(struct.pack("<f", v))
            return cls(type=PropertyType.VECTOR, data=bytes(data))

        if isinstance(value, dict):
            return cls._encode_map(value)

        raise TypeError(f"Unsupported type: {type(value)}")

    def to_python(self) -> Any:
        """Convert TypedValue to a Python value."""
        if self.is_null or self.type == PropertyType.NULL:
            return None

        if self.type == PropertyType.BOOL:
            return self.data[0] != 0 if len(self.data) >= 1 else False

        if self.type == PropertyType.INT32:
            return struct.unpack("<i", self.data)[0] if len(self.data) >= 4 else 0

        if self.type == PropertyType.UINT32:
            return struct.unpack("<I", self.data)[0] if len(self.data) >= 4 else 0

        if self.type == PropertyType.INT64:
            return struct.unpack("<q", self.data)[0] if len(self.data) >= 8 else 0

        if self.type == PropertyType.UINT64:
            return struct.unpack("<Q", self.data)[0] if len(self.data) >= 8 else 0

        if self.type == PropertyType.FLOAT:
            return struct.unpack("<f", self.data)[0] if len(self.data) >= 4 else 0.0

        if self.type == PropertyType.DOUBLE:
            return struct.unpack("<d", self.data)[0] if len(self.data) >= 8 else 0.0

        if self.type in (PropertyType.STRING, PropertyType.TEXT):
            try:
                return self.data.decode("utf-8")
            except UnicodeDecodeError:
                # Database may return binary data as STRING type
                # Return raw bytes in this case
                return self.data

        if self.type == PropertyType.BLOB:
            return self.data

        if self.type == PropertyType.TIMESTAMP:
            if len(self.data) >= 12:
                secs, nanos = struct.unpack("<qI", self.data[0:12])
                return datetime.fromtimestamp(secs, tz=timezone.utc).replace(
                    microsecond=nanos // 1000)
            return datetime.min.replace(tzinfo=timezone.utc)

        if self.type == PropertyType.DATE:
            if len(self.data) >= 4:
                year = struct.unpack("<H", self.data[0:2])[0]
                month = self.data[2]
                day = self.data[3]
                return GqldbDate(year=year, month=month, day=day)
            return GqldbDate(year=1970, month=1, day=1)

        if self.type == PropertyType.LIST:
            return self._decode_list()

        if self.type == PropertyType.MAP:
            return self._decode_map()

        if self.type == PropertyType.POINT:
            if len(self.data) >= 16:
                lng, lat = struct.unpack("<dd", self.data[:16])
                return Point(latitude=lat, longitude=lng)
            return Point(latitude=0.0, longitude=0.0)

        if self.type == PropertyType.POINT3D:
            if len(self.data) >= 24:
                x, y, z = struct.unpack("<ddd", self.data[:24])
                return Point3D(x=x, y=y, z=z)
            return Point3D(x=0.0, y=0.0, z=0.0)

        if self.type == PropertyType.DECIMAL:
            return GqldbDecimal(value=self.data.decode("utf-8"))

        if self.type == PropertyType.SET:
            return set(self._decode_list())

        if self.type in (PropertyType.LOCAL_DATETIME, PropertyType.DATETIME):
            if len(self.data) >= 11:
                year, month, day, hour, minute, second = struct.unpack("<HBBBBB", self.data[0:7])
                nanosecond = struct.unpack("<I", self.data[7:11])[0]
                return GqldbLocalDateTime(year=year, month=month, day=day,
                                          hour=hour, minute=minute, second=second,
                                          nanosecond=nanosecond)
            return GqldbLocalDateTime(year=1970, month=1, day=1, hour=0, minute=0, second=0, nanosecond=0)

        if self.type == PropertyType.ZONED_DATETIME:
            if len(self.data) >= 13:
                year, month, day, hour, minute, second = struct.unpack("<HBBBBB", self.data[0:7])
                nanosecond = struct.unpack("<I", self.data[7:11])[0]
                offset_minutes = struct.unpack("<h", self.data[11:13])[0]
                return GqldbZonedDateTime(year=year, month=month, day=day,
                                          hour=hour, minute=minute, second=second,
                                          nanosecond=nanosecond, offset_minutes=offset_minutes)
            return GqldbZonedDateTime(year=1970, month=1, day=1, hour=0, minute=0, second=0,
                                      nanosecond=0, offset_minutes=0)

        if self.type == PropertyType.LOCAL_TIME:
            if len(self.data) >= 8:
                hour = self.data[0]
                minute = self.data[1]
                second = self.data[2]
                nanosecond = struct.unpack("<I", self.data[4:8])[0]
                return GqldbLocalTime(hour=hour, minute=minute, second=second, nanosecond=nanosecond)
            return GqldbLocalTime(hour=0, minute=0, second=0, nanosecond=0)

        if self.type == PropertyType.ZONED_TIME:
            if len(self.data) >= 10:
                hour = self.data[0]
                minute = self.data[1]
                second = self.data[2]
                nanosecond = struct.unpack("<I", self.data[4:8])[0]
                offset_minutes = struct.unpack("<h", self.data[8:10])[0]
                return GqldbZonedTime(hour=hour, minute=minute, second=second, nanosecond=nanosecond, offset_minutes=offset_minutes)
            return GqldbZonedTime(hour=0, minute=0, second=0, nanosecond=0, offset_minutes=0)

        if self.type == PropertyType.YEAR_TO_MONTH:
            if len(self.data) >= 4:
                months = struct.unpack("<i", self.data[:4])[0]
                return YearToMonth(months=months)
            return YearToMonth(months=0)

        if self.type == PropertyType.DAY_TO_SECOND:
            if len(self.data) >= 12:
                seconds = struct.unpack("<Q", self.data[0:8])[0]
                nanoseconds = struct.unpack("<I", self.data[8:12])[0]
                return DayToSecond(seconds=seconds, nanoseconds=nanoseconds)
            return DayToSecond(seconds=0, nanoseconds=0)

        if self.type == PropertyType.VECTOR:
            if len(self.data) >= 4:
                dim = struct.unpack("<I", self.data[:4])[0]
                if len(self.data) >= 4 + dim * 4:
                    values = []
                    for i in range(dim):
                        v = struct.unpack("<f", self.data[4 + i*4:8 + i*4])[0]
                        values.append(v)
                    return Vector(values=values)
            return Vector(values=[])

        if self.type == PropertyType.RECORD:
            return self._decode_map()

        if self.type == PropertyType.TABLE:
            return self._decode_table()

        if self.type == PropertyType.PATH:
            return self._decode_path()

        if self.type == PropertyType.ERROR:
            return self._decode_error()

        if self.type == PropertyType.NODE:
            return self._decode_node()

        if self.type == PropertyType.EDGE:
            return self._decode_edge()

        # For unsupported types, return raw bytes
        return self.data

    def format_value(self) -> str:
        """Serialize TypedValue to canonical string representation."""
        if self.is_null or self.type == PropertyType.NULL:
            return ""

        val = self.to_python()

        if self.type == PropertyType.BOOL:
            return "true" if val else "false"

        if self.type in (PropertyType.INT32, PropertyType.UINT32, PropertyType.INT64,
                         PropertyType.UINT64, PropertyType.FLOAT, PropertyType.DOUBLE):
            return str(val)

        if self.type in (PropertyType.STRING, PropertyType.TEXT):
            return val if isinstance(val, str) else val.decode("utf-8")

        if self.type == PropertyType.DECIMAL:
            return val.value

        if self.type == PropertyType.DATE:
            return f"{val.year:04d}-{val.month:02d}-{val.day:02d}"

        if self.type == PropertyType.LOCAL_TIME:
            return _format_time_component(val.hour, val.minute, val.second, val.nanosecond)

        if self.type == PropertyType.ZONED_TIME:
            t = _format_time_component(val.hour, val.minute, val.second, val.nanosecond)
            return t + _format_offset(val.offset_minutes)

        if self.type in (PropertyType.LOCAL_DATETIME, PropertyType.DATETIME):
            d = f"{val.year:04d}-{val.month:02d}-{val.day:02d}"
            t = _format_time_component(val.hour, val.minute, val.second, val.nanosecond)
            return d + " " + t

        if self.type == PropertyType.ZONED_DATETIME:
            d = f"{val.year:04d}-{val.month:02d}-{val.day:02d}"
            t = _format_time_component(val.hour, val.minute, val.second, val.nanosecond)
            return d + " " + t + _format_offset(val.offset_minutes)

        if self.type == PropertyType.TIMESTAMP:
            # val is a datetime in UTC
            nanos = val.microsecond * 1000
            d = f"{val.year:04d}-{val.month:02d}-{val.day:02d}"
            t = _format_time_component(val.hour, val.minute, val.second, nanos)
            return d + " " + t

        if self.type == PropertyType.YEAR_TO_MONTH:
            months = val.months
            if months == 0:
                return "P0M"
            neg = months < 0
            abs_months = abs(months)
            years = abs_months // 12
            rem = abs_months % 12
            s = "-P" if neg else "P"
            if years > 0:
                s += f"{years}Y"
            if rem > 0:
                s += f"{rem}M"
            return s

        if self.type == PropertyType.DAY_TO_SECOND:
            return _format_day_to_second(val.seconds, val.nanoseconds)

        # Fallback
        return str(val)

    @staticmethod
    def from_string(s: str, target_type: PropertyType) -> "TypedValue":
        """Parse a string into a TypedValue for the given target type."""
        if s == "":
            return TypedValue(type=target_type, data=b"", is_null=True)

        if target_type == PropertyType.NULL:
            return TypedValue(type=PropertyType.NULL, data=b"", is_null=True)

        if target_type == PropertyType.BOOL:
            val = s.lower() in ("true", "1")
            return TypedValue.from_value(val)

        if target_type == PropertyType.INT32:
            return TypedValue(type=PropertyType.INT32, data=struct.pack("<i", int(s)))

        if target_type == PropertyType.UINT32:
            return TypedValue(type=PropertyType.UINT32, data=struct.pack("<I", int(s)))

        if target_type == PropertyType.INT64:
            return TypedValue(type=PropertyType.INT64, data=struct.pack("<q", int(s)))

        if target_type == PropertyType.UINT64:
            return TypedValue(type=PropertyType.UINT64, data=struct.pack("<Q", int(s)))

        if target_type == PropertyType.FLOAT:
            return TypedValue(type=PropertyType.FLOAT, data=struct.pack("<f", float(s)))

        if target_type == PropertyType.DOUBLE:
            return TypedValue(type=PropertyType.DOUBLE, data=struct.pack("<d", float(s)))

        if target_type in (PropertyType.STRING, PropertyType.TEXT):
            return TypedValue(type=target_type, data=s.encode("utf-8"))

        if target_type == PropertyType.DECIMAL:
            return TypedValue.from_value(GqldbDecimal(value=s))

        if target_type == PropertyType.DATE:
            parts = s.split("-")
            year, month, day = int(parts[0]), int(parts[1]), int(parts[2])
            return TypedValue.from_value(GqldbDate(year=year, month=month, day=day))

        if target_type == PropertyType.LOCAL_TIME:
            clean = _remove_offset(s)
            hour, minute, second, nanos = _parse_time_string(clean)
            return TypedValue.from_value(GqldbLocalTime(hour=hour, minute=minute, second=second, nanosecond=nanos))

        if target_type == PropertyType.ZONED_TIME:
            time_part, offset_minutes = _split_time_and_offset(s)
            hour, minute, second, nanos = _parse_time_string(time_part)
            return TypedValue.from_value(GqldbZonedTime(hour=hour, minute=minute, second=second,
                                                         nanosecond=nanos, offset_minutes=offset_minutes))

        if target_type in (PropertyType.LOCAL_DATETIME, PropertyType.DATETIME):
            clean = _remove_offset(s)
            sep = "T" if "T" in clean else " "
            date_part, time_part = clean.split(sep, 1)
            dp = date_part.split("-")
            year, month, day = int(dp[0]), int(dp[1]), int(dp[2])
            hour, minute, second, nanos = _parse_time_string(time_part)
            return TypedValue.from_value(GqldbLocalDateTime(year=year, month=month, day=day,
                                                             hour=hour, minute=minute, second=second,
                                                             nanosecond=nanos))

        if target_type == PropertyType.ZONED_DATETIME:
            normalized = _normalize_offset(s)
            sep = "T" if "T" in normalized else " "
            date_part, rest = normalized.split(sep, 1)
            time_part, offset_minutes = _split_time_and_offset(rest)
            dp = date_part.split("-")
            year, month, day = int(dp[0]), int(dp[1]), int(dp[2])
            hour, minute, second, nanos = _parse_time_string(time_part)
            return TypedValue.from_value(GqldbZonedDateTime(year=year, month=month, day=day,
                                                             hour=hour, minute=minute, second=second,
                                                             nanosecond=nanos, offset_minutes=offset_minutes))

        if target_type == PropertyType.TIMESTAMP:
            # Support epoch seconds as string
            if s.isdigit() or (s.startswith("-") and s[1:].isdigit()):
                epoch = int(s)
                data = struct.pack("<qI", epoch, 0)
                return TypedValue(type=PropertyType.TIMESTAMP, data=data)
            # Handle Z suffix (treat as UTC, +00:00)
            if s.endswith("Z"):
                s = s[:-1] + "+00:00"
            # Check for timezone offset and convert to UTC if present
            # Allow optional space before offset
            tz_match = re.search(r'\s?([+-])(\d{2}):?(\d{2})$', s)
            if tz_match:
                dt_part = s[:tz_match.start()]
                sign = 1 if tz_match.group(1) == "+" else -1
                oh = int(tz_match.group(2))
                om = int(tz_match.group(3))
                offset_minutes = sign * (oh * 60 + om)
                # Split date and time from the datetime portion
                sep = "T" if "T" in dt_part else " "
                date_part, time_part = dt_part.split(sep, 1)
                dp = date_part.split("-")
                year, month, day = int(dp[0]), int(dp[1]), int(dp[2])
                hour, minute, second, nanos = _parse_time_string(time_part)
                # Build a datetime with the offset, then convert to UTC
                tz_offset = timezone(timedelta(minutes=offset_minutes))
                local_dt = datetime(year, month, day, hour, minute, second,
                                    nanos // 1000, tzinfo=tz_offset)
                utc_dt = local_dt.astimezone(timezone.utc)
                utc_secs = int(utc_dt.replace(tzinfo=timezone.utc).timestamp())
                data = struct.pack("<qI", utc_secs, nanos)
                return TypedValue(type=PropertyType.TIMESTAMP, data=data)
            # Otherwise parse "yyyy-MM-dd HH:mm:ss[.n]" (no timezone, treated as UTC)
            sep = "T" if "T" in s else " "
            date_part, time_part = s.split(sep, 1)
            dp = date_part.split("-")
            year, month, day = int(dp[0]), int(dp[1]), int(dp[2])
            hour, minute, second, nanos = _parse_time_string(time_part)
            secs = calendar.timegm((year, month, day, hour, minute, second, 0, 0, 0))
            data = struct.pack("<qI", secs, nanos)
            return TypedValue(type=PropertyType.TIMESTAMP, data=data)

        if target_type == PropertyType.YEAR_TO_MONTH:
            months = _parse_year_to_month(s)
            return TypedValue.from_value(YearToMonth(months=months))

        if target_type == PropertyType.DAY_TO_SECOND:
            secs, nanos = _parse_day_to_second(s)
            return TypedValue.from_value(DayToSecond(seconds=secs, nanoseconds=nanos))

        raise ValueError(f"Unsupported target type for from_string: {target_type}")

    def _decode_path(self) -> "GqldbPath":
        """Decode a path from binary format.

        Binary format: [nodeCount:2][nodeLen:4][nodeData]...[edgeCount:2][edgeLen:4][edgeData]...
        """
        offset = 0
        nodes = []
        edges = []

        try:
            # Decode nodes
            if offset + 2 > len(self.data):
                return GqldbPath(nodes=nodes, edges=edges)
            node_count = struct.unpack("<H", self.data[offset:offset+2])[0]
            offset += 2

            for _ in range(node_count):
                if offset + 4 > len(self.data):
                    break
                node_len = struct.unpack("<I", self.data[offset:offset+4])[0]
                offset += 4
                if offset + node_len > len(self.data):
                    break
                node_data = self.data[offset:offset+node_len]
                nodes.append(self._decode_node_binary(node_data))
                offset += node_len

            # Decode edges
            if offset + 2 > len(self.data):
                return GqldbPath(nodes=nodes, edges=edges)
            edge_count = struct.unpack("<H", self.data[offset:offset+2])[0]
            offset += 2

            for _ in range(edge_count):
                if offset + 4 > len(self.data):
                    break
                edge_len = struct.unpack("<I", self.data[offset:offset+4])[0]
                offset += 4
                if offset + edge_len > len(self.data):
                    break
                edge_data = self.data[offset:offset+edge_len]
                edges.append(self._decode_edge_binary(edge_data))
                offset += edge_len

            return GqldbPath(nodes=nodes, edges=edges)
        except Exception:
            return GqldbPath()

    def _decode_error(self) -> "GqldbError":
        """Decode an error from JSON bytes."""
        try:
            json_str = self.data.decode("utf-8")
            err_data = json.loads(json_str)
            return GqldbError(
                code=err_data.get("code", 0),
                message=err_data.get("message", json_str)
            )
        except Exception:
            # If not valid JSON, treat as plain message
            return GqldbError(message=self.data.decode("utf-8"))

    def _decode_node(self) -> "GqldbNode":
        """Decode a node from binary format.

        Binary format: [idLen:2][id][labelCount:2][labelLen:2][label]...[properties_binary]
        """
        return self._decode_node_binary(self.data)

    def _decode_node_binary(self, data: bytes) -> "GqldbNode":
        """Decode a node from binary data."""
        offset = 0
        try:
            # Decode ID
            node_id, id_len = self._decode_string_from(data, offset)
            if id_len == 0:
                return GqldbNode()
            offset += id_len

            # Decode labels
            if offset + 2 > len(data):
                return GqldbNode(id=node_id)
            label_count = struct.unpack("<H", data[offset:offset+2])[0]
            offset += 2

            labels = []
            for _ in range(label_count):
                label, label_len = self._decode_string_from(data, offset)
                if label_len == 0:
                    break
                labels.append(label)
                offset += label_len

            # Decode properties
            properties = self._decode_properties_binary(data, offset)

            return GqldbNode(id=node_id, labels=labels, properties=properties)
        except Exception:
            return GqldbNode()

    def _decode_edge(self) -> "GqldbEdge":
        """Decode an edge from binary format.

        Binary format: [idLen:2][id][labelLen:2][label][fromLen:2][from][toLen:2][to][properties_binary]
        """
        return self._decode_edge_binary(self.data)

    def _decode_edge_binary(self, data: bytes) -> "GqldbEdge":
        """Decode an edge from binary data."""
        offset = 0
        try:
            # Decode ID
            edge_id, id_len = self._decode_string_from(data, offset)
            if id_len == 0:
                return GqldbEdge()
            offset += id_len

            # Decode label
            label, label_len = self._decode_string_from(data, offset)
            if label_len == 0:
                return GqldbEdge(id=edge_id)
            offset += label_len

            # Decode from_node_id
            from_id, from_len = self._decode_string_from(data, offset)
            if from_len == 0:
                return GqldbEdge(id=edge_id, label=label)
            offset += from_len

            # Decode to_node_id
            to_id, to_len = self._decode_string_from(data, offset)
            if to_len == 0:
                return GqldbEdge(id=edge_id, label=label, from_node_id=from_id)
            offset += to_len

            # Decode properties
            properties = self._decode_properties_binary(data, offset)

            return GqldbEdge(id=edge_id, label=label, from_node_id=from_id, to_node_id=to_id, properties=properties)
        except Exception:
            return GqldbEdge()

    def _decode_table(self) -> "GqldbTable":
        """Decode a table from binary format.

        Binary format: [colCount:2][colLen:2][col]...[rowCount:2][cellCount:2][TVE]...
        """
        offset = 0
        columns = []
        rows = []
        column_types = []

        try:
            # Decode columns
            if offset + 2 > len(self.data):
                return GqldbTable()
            col_count = struct.unpack("<H", self.data[offset:offset+2])[0]
            offset += 2

            for _ in range(col_count):
                col, col_len = self._decode_string_from(self.data, offset)
                if col_len == 0:
                    break
                columns.append(col)
                offset += col_len

            # Decode rows
            if offset + 2 > len(self.data):
                return GqldbTable(columns=columns)
            row_count = struct.unpack("<H", self.data[offset:offset+2])[0]
            offset += 2

            for row_idx in range(row_count):
                if offset + 2 > len(self.data):
                    break
                cell_count = struct.unpack("<H", self.data[offset:offset+2])[0]
                offset += 2

                row = []
                for cell_idx in range(cell_count):
                    tv, tv_len = self._decode_tve_from(self.data, offset)
                    if tv_len == 0:
                        break
                    row.append(tv.to_python())
                    # Capture column types from first row
                    if row_idx == 0:
                        column_types.append(tv.type)
                    offset += tv_len
                rows.append(row)

            return GqldbTable(columns=columns, rows=rows, column_types=column_types)
        except Exception:
            return GqldbTable()

    def _decode_string_from(self, data: bytes, offset: int) -> tuple:
        """Decode a length-prefixed string: [len:2][utf8 bytes]. Returns (string, bytes consumed)."""
        if offset + 2 > len(data):
            return "", 0
        length = struct.unpack("<H", data[offset:offset+2])[0]
        if offset + 2 + length > len(data):
            return "", 0
        return data[offset+2:offset+2+length].decode("utf-8"), 2 + length

    def _decode_tve_from(self, data: bytes, offset: int) -> tuple:
        """Decode a TypedValueEntry: [type:1][isNull:1][dataLen:4][data]. Returns (TypedValue, bytes consumed)."""
        if offset + 6 > len(data):
            return TypedValue(type=PropertyType.NULL, data=b"", is_null=True), 0
        tv_type = PropertyType(data[offset])
        is_null = data[offset+1] != 0
        data_len = struct.unpack("<I", data[offset+2:offset+6])[0]
        if offset + 6 + data_len > len(data):
            return TypedValue(type=PropertyType.NULL, data=b"", is_null=True), 0
        tv_data = data[offset+6:offset+6+data_len]
        return TypedValue(type=tv_type, data=tv_data, is_null=is_null), 6 + data_len

    def _decode_properties_binary(self, data: bytes, offset: int) -> Dict[str, Any]:
        """Decode binary properties: [count:2][keyLen:2][key][TVE]..."""
        result = {}
        if offset + 2 > len(data):
            return result
        count = struct.unpack("<H", data[offset:offset+2])[0]
        pos = offset + 2

        for _ in range(count):
            key, key_len = self._decode_string_from(data, pos)
            if key_len == 0:
                break
            pos += key_len

            tv, tv_len = self._decode_tve_from(data, pos)
            if tv_len == 0:
                break
            pos += tv_len

            result[key] = tv.to_python()

        return result

    @classmethod
    def _encode_list(cls, values: List[Any]) -> "TypedValue":
        """Encode a list to bytes. Binary format: [count:2][TVE]..."""
        data = bytearray()
        # Write count (2 bytes)
        data.extend(struct.pack("<H", len(values)))

        for v in values:
            tv = cls.from_value(v)
            # Write TypedValueEntry: [type:1][isNull:1][dataLen:4][data]
            data.append(tv.type.value)
            data.append(1 if tv.is_null else 0)
            data.extend(struct.pack("<I", len(tv.data)))
            data.extend(tv.data)

        return cls(type=PropertyType.LIST, data=bytes(data))

    def _decode_list(self) -> List[Any]:
        """Decode a list from bytes.

        Supports both JSON format (e.g., "[1,2,3]") and binary format.
        Binary format: [count:2][TVE]... where TVE is [type:1][isNull:1][dataLen:4][data]
        """
        if len(self.data) == 0:
            return []

        # Check if data is JSON format (starts with '[')
        if self.data[0:1] == b'[':
            try:
                return json.loads(self.data.decode('utf-8'))
            except (json.JSONDecodeError, UnicodeDecodeError):
                return []

        # Binary format: [count:2][TVE]...
        if len(self.data) < 2:
            return []

        count = struct.unpack("<H", self.data[:2])[0]
        offset = 2
        result = []

        for _ in range(count):
            tv, tv_len = self._decode_tve_from(self.data, offset)
            if tv_len == 0:
                break
            offset += tv_len
            result.append(tv.to_python())

        return result

    @classmethod
    def _encode_map(cls, m: Dict[str, Any]) -> "TypedValue":
        """Encode a map to bytes. Binary format: [count:2][keyLen:2][key][TVE]..."""
        data = bytearray()
        # Write count (2 bytes)
        data.extend(struct.pack("<H", len(m)))

        for k, v in m.items():
            # Write key (2-byte length prefix)
            key_bytes = k.encode("utf-8")
            data.extend(struct.pack("<H", len(key_bytes)))
            data.extend(key_bytes)

            # Write value as TVE
            tv = cls.from_value(v)
            data.append(tv.type.value)
            data.append(1 if tv.is_null else 0)
            data.extend(struct.pack("<I", len(tv.data)))
            data.extend(tv.data)

        return cls(type=PropertyType.MAP, data=bytes(data))

    @classmethod
    def _encode_set(cls, values: set) -> "TypedValue":
        """Encode a set to bytes. Binary format: [count:2][TVE]..."""
        data = bytearray()
        # Write count (2 bytes)
        data.extend(struct.pack("<H", len(values)))

        for v in values:
            tv = cls.from_value(v)
            # Write TypedValueEntry
            data.append(tv.type.value)
            data.append(1 if tv.is_null else 0)
            data.extend(struct.pack("<I", len(tv.data)))
            data.extend(tv.data)

        return cls(type=PropertyType.SET, data=bytes(data))

    def _decode_map(self) -> Dict[str, Any]:
        """Decode a map from bytes. Binary format: [count:2][keyLen:2][key][TVE]..."""
        if len(self.data) < 2:
            return {}

        count = struct.unpack("<H", self.data[:2])[0]
        offset = 2
        result = {}

        for _ in range(count):
            # Read key
            key, key_len = self._decode_string_from(self.data, offset)
            if key_len == 0:
                break
            offset += key_len

            # Read value as TVE
            tv, tv_len = self._decode_tve_from(self.data, offset)
            if tv_len == 0:
                break
            offset += tv_len

            result[key] = tv.to_python()

        return result



@dataclass
class Parameter:
    """Named parameter for queries."""
    name: str
    value: TypedValue

    @classmethod
    def create(cls, name: str, value: Any) -> "Parameter":
        """Create a parameter from name and Python value."""
        return cls(name=name, value=TypedValue.from_value(value))


# ---------------------------------------------------------------------------
# Module-level helper functions for format_value / from_string
# ---------------------------------------------------------------------------

def _format_nanos(nanos: int) -> str:
    """Format nanoseconds as a fractional string with trailing zeros trimmed.
    Returns empty string when nanos == 0."""
    if nanos == 0:
        return ""
    s = f"{nanos:09d}"
    return "." + s.rstrip("0")


def _format_time_component(hour: int, minute: int, second: int, nanos: int) -> str:
    """Format HH:mm:ss[.nnnnnnnnn]."""
    return f"{hour:02d}:{minute:02d}:{second:02d}{_format_nanos(nanos)}"


def _format_offset(offset_minutes: int) -> str:
    """Format offset in minutes as +HH:MM or -HH:MM."""
    if offset_minutes < 0:
        sign = "-"
        total = -offset_minutes
    else:
        sign = "+"
        total = offset_minutes
    h = total // 60
    m = total % 60
    return f"{sign}{h:02d}:{m:02d}"


def _parse_time_string(s: str) -> tuple:
    """Parse HH:mm:ss[.nnnnnnnnn] and return (hour, minute, second, nanos)."""
    dot_idx = s.find(".")
    if dot_idx >= 0:
        hms = s[:dot_idx]
        frac_str = s[dot_idx + 1:]
        # Pad to 9 digits
        frac_str = frac_str.ljust(9, "0")[:9]
        nanos = int(frac_str)
    else:
        hms = s
        nanos = 0
    parts = hms.split(":")
    return int(parts[0]), int(parts[1]), int(parts[2]), nanos


def _remove_offset(s: str) -> str:
    """Strip trailing Z or timezone offset (with optional space) from a time/datetime string.
    Returns the clean local time/datetime string.
    Only inspects the last ~10 characters to avoid matching date hyphens like '1993-05-06'."""
    # Handle trailing Z
    if s.endswith("Z"):
        return s[:-1]
    # Look for offset pattern in the tail of the string
    tail = s[-10:] if len(s) > 10 else s
    m = re.search(r'\s?[+-]\d{2}:\d{2}$', tail)
    if m:
        # Compute position in the full string
        cut = len(s) - (len(tail) - m.start())
        return s[:cut]
    return s


def _normalize_offset(s: str) -> str:
    """Collapse optional space before offset and convert trailing Z to +00:00.
    E.g. '14:30:00 +08:00' -> '14:30:00+08:00', '14:30:00Z' -> '14:30:00+00:00'."""
    if s.endswith("Z"):
        return s[:-1] + "+00:00"
    # Look for space before offset in the tail
    tail = s[-10:] if len(s) > 10 else s
    m = re.search(r'\s([+-]\d{2}:\d{2})$', tail)
    if m:
        cut = len(s) - (len(tail) - m.start())
        return s[:cut] + m.group(1)
    return s


def _split_time_and_offset(s: str) -> tuple:
    """Split a time string with offset like '09:11:02-08:00' or '09:11:02+05:30'.
    Also supports 'Z' suffix and space-separated offsets.
    Returns (time_part, offset_minutes)."""
    # Normalize: collapse space before offset, convert Z to +00:00
    s = _normalize_offset(s)
    # Find the last +/- that indicates offset (skip the first character to handle negative)
    # Offset pattern: [+-]HH:MM at the end
    m = re.search(r'([+-])(\d{2}):(\d{2})$', s)
    if m:
        time_part = s[:m.start()]
        sign = 1 if m.group(1) == "+" else -1
        oh = int(m.group(2))
        om = int(m.group(3))
        offset_minutes = sign * (oh * 60 + om)
        return time_part, offset_minutes
    return s, 0


def _parse_year_to_month(s: str) -> int:
    """Parse ISO-8601 year-to-month duration like 'P2Y5M', 'P0M', '-P1Y5M'."""
    neg = s.startswith("-")
    p = s.lstrip("-")
    if not p.startswith("P"):
        raise ValueError(f"Invalid YEAR_TO_MONTH format: {s}")
    p = p[1:]  # Remove 'P'
    years = 0
    months_part = 0
    ym = re.match(r'(\d+)Y', p)
    if ym:
        years = int(ym.group(1))
        p = p[ym.end():]
    mm = re.match(r'(\d+)M', p)
    if mm:
        months_part = int(mm.group(1))
    total = years * 12 + months_part
    return -total if neg else total


def _parse_day_to_second(s: str) -> tuple:
    """Parse ISO-8601 day-to-second duration like 'P3DT4H', 'PT0S', 'P1DT2H3M4.12S'.
    Returns (total_seconds, nanoseconds)."""
    neg = s.startswith("-")
    p = s.lstrip("-")
    if not p.startswith("P"):
        raise ValueError(f"Invalid DAY_TO_SECOND format: {s}")
    p = p[1:]  # Remove 'P'

    days = 0
    dm = re.match(r'(\d+)D', p)
    if dm:
        days = int(dm.group(1))
        p = p[dm.end():]

    hours = 0
    minutes = 0
    secs = 0
    nanos = 0

    if p.startswith("T"):
        p = p[1:]
        hm = re.match(r'(\d+)H', p)
        if hm:
            hours = int(hm.group(1))
            p = p[hm.end():]
        mm = re.match(r'(\d+)M', p)
        if mm:
            minutes = int(mm.group(1))
            p = p[mm.end():]
        sm = re.match(r'(\d+(?:\.\d+)?)S', p)
        if sm:
            sec_str = sm.group(1)
            if "." in sec_str:
                whole, frac = sec_str.split(".")
                secs = int(whole)
                frac = frac.ljust(9, "0")[:9]
                nanos = int(frac)
            else:
                secs = int(sec_str)

    total_secs = days * 86400 + hours * 3600 + minutes * 60 + secs
    if neg:
        total_secs = -total_secs
        # nanos stays positive per the DayToSecond encoding
    return total_secs, nanos


def _format_day_to_second(seconds: int, nanoseconds: int) -> str:
    """Format DayToSecond as ISO-8601 duration."""
    if seconds == 0 and nanoseconds == 0:
        return "PT0S"

    neg = seconds < 0
    abs_secs = abs(seconds)

    days = abs_secs // 86400
    rem = abs_secs % 86400
    hours = rem // 3600
    rem = rem % 3600
    mins = rem // 60
    secs = rem % 60

    result = "-P" if neg else "P"
    if days > 0:
        result += f"{days}D"

    has_time = hours > 0 or mins > 0 or secs > 0 or nanoseconds > 0
    if has_time:
        result += "T"
        if hours > 0:
            result += f"{hours}H"
        if mins > 0:
            result += f"{mins}M"
        if secs > 0 or nanoseconds > 0:
            if nanoseconds > 0:
                frac = f"{nanoseconds:09d}".rstrip("0")
                result += f"{secs}.{frac}S"
            else:
                result += f"{secs}S"

    return result



"""Integration tests for useGraph() functionality."""

import pytest
import time

from gqldb.config import ConfigBuilder
from gqldb.client import GqldbClient
from tests.conftest import TEST_HOST, TEST_USERNAME, TEST_PASSWORD, TEST_DATABASE


class TestUseGraphIntegration:
    """Integration tests for useGraph functionality."""

    @pytest.fixture
    def client(self):
        """Create a test client."""
        config = (
            ConfigBuilder()
            .hosts(TEST_HOST)
            .username(TEST_USERNAME)
            .password(TEST_PASSWORD)
            .default_graph(TEST_DATABASE)
            .timeout(30)
            .build()
        )
        client = GqldbClient(config)
        try:
            client.login(TEST_USERNAME, TEST_PASSWORD)
        except Exception as e:
            if "authentication is not enabled" in str(e).lower():
                pass  # Continue without login
            else:
                raise
        yield client
        try:
            if client.is_logged_in():
                client.logout()
        except Exception:
            pass
        client.close()

    def test_usegraph_changes_default_graph(self, client):
        """useGraph should change the default graph for queries."""
        # Use a known graph
        client.use_graph(TEST_DATABASE)

        # Execute query - should use the graph set by useGraph
        response = client.gql("MATCH (n) RETURN n LIMIT 1")

        assert response is not None
        print(f"useGraph test passed, rows returned: {response.row_count}")

    def test_usegraph_with_nonexistent_graph_error(self, client):
        """useGraph with non-existent graph should cause error."""
        fake_graph = f"non_existent_graph_{int(time.time())}"

        with pytest.raises(Exception) as exc_info:
            client.use_graph(fake_graph)

        error_msg = str(exc_info.value).lower()
        assert "not found" in error_msg or "does not exist" in error_msg or "invalid" in error_msg
        print(f"Got expected error for non-existent graph: {exc_info.value}")

    def test_query_config_overrides_usegraph(self, client):
        """Explicit graphName in QueryConfig should override useGraph."""
        from gqldb.client import QueryConfig

        # Set default graph via useGraph
        client.use_graph(TEST_DATABASE)

        # Query with different graph in config (non-existent)
        fake_graph = f"override_test_graph_{int(time.time())}"

        with pytest.raises(Exception) as exc_info:
            client.gql("MATCH (n) RETURN n LIMIT 1", QueryConfig(graph_name=fake_graph))

        error_msg = str(exc_info.value).lower()
        assert "not found" in error_msg or "does not exist" in error_msg or "invalid" in error_msg
        print(f"QueryConfig.graph_name correctly overrides useGraph: {exc_info.value}")

    def test_usegraph_persists_across_queries(self, client):
        """useGraph should persist across multiple queries."""
        client.use_graph(TEST_DATABASE)

        # Execute multiple queries
        for i in range(3):
            response = client.gql("MATCH (n) RETURN n LIMIT 1")
            assert response is not None

        print("useGraph persisted across 3 queries")

    def test_usegraph_overrides_config_default_graph(self):
        """useGraph should override config.default_graph."""
        # Create a new client with no default graph in config
        config = (
            ConfigBuilder()
            .hosts(TEST_HOST)
            .username(TEST_USERNAME)
            .password(TEST_PASSWORD)
            .default_graph("")  # No default graph
            .timeout(30)
            .build()
        )

        test_client = GqldbClient(config)

        try:
            try:
                test_client.login(TEST_USERNAME, TEST_PASSWORD)
            except Exception as e:
                if "authentication is not enabled" not in str(e).lower():
                    raise

            # Use a valid graph via useGraph
            test_client.use_graph(TEST_DATABASE)

            # Query should succeed because useGraph sets session.default_graph
            response = test_client.gql("MATCH (n) RETURN n LIMIT 1")
            assert response is not None
            print("useGraph successfully set session.default_graph")
        finally:
            test_client.close()

    def test_usegraph_empty_string_rejected(self, client):
        """useGraph with empty string should be rejected by server."""
        with pytest.raises(Exception) as exc_info:
            client.use_graph("")

        error_msg = str(exc_info.value).lower()
        assert "required" in error_msg or "invalid" in error_msg or "empty" in error_msg
        print(f"Server correctly rejected empty graph name: {exc_info.value}")

    def test_multiple_usegraph_uses_latest(self, client):
        """Multiple useGraph calls should use the latest value."""
        # First useGraph with valid graph
        client.use_graph(TEST_DATABASE)

        # Second useGraph with same graph
        client.use_graph(TEST_DATABASE)

        # Query should succeed
        response = client.gql("MATCH (n) RETURN n LIMIT 1")
        assert response is not None
        print("Latest useGraph value is correctly used")

    def test_usegraph_with_gql_stream(self, client):
        """useGraph should work with gql_stream."""
        client.use_graph(TEST_DATABASE)

        responses = []
        client.gql_stream("MATCH (n) RETURN n LIMIT 5", callback=lambda r: responses.append(r))

        assert len(responses) > 0
        print(f"gql_stream received {len(responses)} responses with useGraph")



"""Graph Service integration tests."""

import pytest
import time

from gqldb.client import GqldbClient
from gqldb.types import GraphType
from tests.conftest import drop_test_graph


@pytest.mark.integration
class TestGraph:
    """Tests for graph service."""

    def test_list_graphs(self, connected_client: GqldbClient):
        """Test listing available graphs."""
        graphs = connected_client.list_graphs()
        assert isinstance(graphs, list)
        print(f"Found {len(graphs)} graphs:")
        for g in graphs:
            print(f"  - {g.name} (type={g.graph_type}, nodes={g.node_count}, edges={g.edge_count})")

    def test_create_and_drop_graph(self, connected_client: GqldbClient):
        """Test creating and dropping a graph."""
        graph_name = f"test_graph_{int(time.time())}"

        # Create graph
        connected_client.create_graph(graph_name, GraphType.OPEN, "Test graph")
        print(f"Created graph: {graph_name}")

        # Verify graph exists
        graphs = connected_client.list_graphs()
        graph_names = [g.name for g in graphs]
        assert graph_name in graph_names

        # Get graph info
        info = connected_client.get_graph_info(graph_name)
        assert info.name == graph_name
        assert info.graph_type == GraphType.OPEN
        print(f"Graph info: {info}")

        # Drop graph (switch away first)
        drop_test_graph(connected_client, graph_name)
        print(f"Dropped graph: {graph_name}")

        # Verify graph is gone
        graphs = connected_client.list_graphs()
        graph_names = [g.name for g in graphs]
        assert graph_name not in graph_names

    def test_drop_graph_if_exists(self, connected_client: GqldbClient):
        """Test dropping a non-existent graph with if_exists=True."""
        # Should not raise error
        connected_client.drop_graph("non_existent_graph_12345", if_exists=True)

    def test_get_graph_info(self, connected_client: GqldbClient):
        """Test getting graph info for an existing graph."""
        info = connected_client.get_graph_info("miniCircle")
        assert info is not None
        assert info.name == "miniCircle"
        print(f"Graph info for miniCircle: type={info.graph_type}, nodes={info.node_count}, edges={info.edge_count}")



"""Graph service handles graph management operations."""

from __future__ import annotations

from typing import TYPE_CHECKING, List

import grpc

if TYPE_CHECKING:
    from ..types import GraphInfo, GraphType
    from .service_context import ServiceContext


class GraphService:
    """Graph service for managing graphs."""

    def __init__(self, ctx: ServiceContext):
        self.ctx = ctx
        self._stub = None

    def _get_stub(self):
        if self._stub is None:
            from ..proto import gqldb_pb2_grpc
            channel = self.ctx.get_channel()
            self._stub = gqldb_pb2_grpc.GraphServiceStub(channel)
        return self._stub

    def create_graph(self, name: str, graph_type: GraphType, description: str) -> None:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        try:
            request = gqldb_pb2.CreateGraphRequest(
                name=name,
                graph_type=graph_type.value,
                description=description
            )
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().CreateGraph(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            if not response.success:
                raise GqldbError(response.message)
        except grpc.RpcError as e:
            raise GqldbError(f"Create graph failed: {e.details()}") from e

    def drop_graph(self, name: str, if_exists: bool) -> None:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        try:
            request = gqldb_pb2.DropGraphRequest(name=name, if_exists=if_exists)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().DropGraph(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            if not response.success:
                raise GqldbError(response.message)
        except grpc.RpcError as e:
            raise GqldbError(f"Drop graph failed: {e.details()}") from e

    def use_graph(self, name: str) -> None:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        session = self.ctx.sessions.get_session()
        try:
            request = gqldb_pb2.UseGraphRequest(
                name=name, session_id=session.id if session else 0
            )
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().UseGraph(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            if not response.success:
                raise GqldbError(response.message)
            self.ctx.sessions.set_default_graph(name)
        except grpc.RpcError as e:
            raise GqldbError(f"Use graph failed: {e.details()}") from e

    def list_graphs(self) -> List[GraphInfo]:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from .converters import convert_graph_info

        try:
            request = gqldb_pb2.ListGraphsRequest()
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().ListGraphs(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            return [convert_graph_info(g) for g in response.graphs]
        except grpc.RpcError as e:
            raise GqldbError(f"List graphs failed: {e.details()}") from e

    def get_graph_info(self, name: str) -> GraphInfo:
        from ..errors import GqldbError, GraphNotFoundError
        from ..proto import gqldb_pb2
        from .converters import convert_graph_info

        try:
            request = gqldb_pb2.GetGraphInfoRequest(name=name)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().GetGraphInfo(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            if not response.info or not response.info.name:
                raise GraphNotFoundError(name)
            return convert_graph_info(response.info)
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.NOT_FOUND:
                raise GraphNotFoundError(name) from e
            raise GqldbError(f"Get graph info failed: {e.details()}") from e



"""GQLDB Python Driver - A gRPC-based client for GQLDB graph database."""

from .client import GqldbClient
from .config import GqldbConfig
from .response import Response, Row, Node, Edge, Path
from .types import TypedValue, PropertyType, Schema, PropertyDef, Table, Header, Attr
from .printers import (
    print_nodes,
    print_nodes_without_schema,
    print_edges,
    print_edges_without_schema,
    print_paths,
    print_table,
    print_any,
)

__version__ = "0.1.0"
__all__ = [
    "GqldbClient",
    "GqldbConfig",
    "Response",
    "Row",
    "Node",
    "Edge",
    "Path",
    "TypedValue",
    "PropertyType",
    "Schema",
    "PropertyDef",
    "Table",
    "Header",
    "Attr",
    "print_nodes",
    "print_nodes_without_schema",
    "print_edges",
    "print_edges_without_schema",
    "print_paths",
    "print_table",
    "print_any",
]



"""Bulk import service handles bulk import session management."""

from __future__ import annotations

from typing import TYPE_CHECKING

import grpc

if TYPE_CHECKING:
    from ..types import (
        AbortBulkImportResult,
        BulkImportSession,
        BulkImportStatus,
        CheckpointResult,
        EndBulkImportResult,
    )
    from .service_context import ServiceContext


class BulkImportService:
    """Bulk import service for managing bulk import sessions."""

    def __init__(self, ctx: ServiceContext):
        self.ctx = ctx
        self._stub = None

    def _get_stub(self):
        if self._stub is None:
            from ..proto import gqldb_pb2_grpc
            channel = self.ctx.get_channel()
            self._stub = gqldb_pb2_grpc.BulkImportServiceStub(channel)
        return self._stub

    def start_bulk_import(
        self,
        graph_name: str,
        checkpoint_every: int,
        estimated_nodes: int,
        estimated_edges: int,
        memtable_size: int,
        max_memtables: int
    ) -> BulkImportSession:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import BulkImportSession

        try:
            request = gqldb_pb2.StartBulkImportRequest(
                graph_name=graph_name,
                checkpoint_every=checkpoint_every,
                estimated_nodes=estimated_nodes,
                estimated_edges=estimated_edges,
                memtable_size=memtable_size,
                max_memtables=max_memtables
            )
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().StartBulkImport(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            if not response.success:
                raise GqldbError(f"Start bulk import failed: {response.message}")
            return BulkImportSession(
                success=response.success,
                session_id=response.session_id,
                message=response.message
            )
        except grpc.RpcError as e:
            raise GqldbError(f"Start bulk import failed: {e.details()}") from e

    def checkpoint(self, session_id: str) -> CheckpointResult:
        """Deprecated: Checkpoint is no longer needed."""
        from ..types import CheckpointResult
        return CheckpointResult(
            success=True,
            record_count=0,
            last_checkpoint_count=0,
            message="Checkpoint has been removed; use end_bulk_import which performs a final flush",
        )

    def end_bulk_import(self, session_id: str) -> EndBulkImportResult:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import EndBulkImportResult

        try:
            request = gqldb_pb2.EndBulkImportRequest(session_id=session_id)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().EndBulkImport(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            return EndBulkImportResult(
                success=response.success,
                total_records=response.total_records,
                message=response.message
            )
        except grpc.RpcError as e:
            raise GqldbError(f"End bulk import failed: {e.details()}") from e

    def abort_bulk_import(self, session_id: str) -> AbortBulkImportResult:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import AbortBulkImportResult

        try:
            request = gqldb_pb2.AbortBulkImportRequest(session_id=session_id)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().AbortBulkImport(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            return AbortBulkImportResult(
                success=response.success,
                message=response.message
            )
        except grpc.RpcError as e:
            raise GqldbError(f"Abort bulk import failed: {e.details()}") from e

    def get_bulk_import_status(self, session_id: str) -> BulkImportStatus:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import BulkImportStatus

        try:
            request = gqldb_pb2.GetBulkImportStatusRequest(session_id=session_id)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().GetBulkImportStatus(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            return BulkImportStatus(
                is_active=response.is_active,
                graph_name=response.graph_name,
                record_count=response.record_count,
                last_checkpoint_count=response.last_checkpoint_count,
                created_at=response.created_at,
                last_activity=response.last_activity
            )
        except grpc.RpcError as e:
            raise GqldbError(f"Get bulk import status failed: {e.details()}") from e



"""Bulk Import Service integration tests."""

import pytest
import uuid

from gqldb.client import GqldbClient
from gqldb.types import GraphType, NodeData, EdgeData
from tests.conftest import drop_test_graph


@pytest.mark.integration
class TestBulkImport:
    """Tests for bulk import service."""

    @pytest.fixture
    def test_graph(self, connected_client: GqldbClient, request):
        """Create a temporary test graph."""
        graph_name = f"test_{request.node.name}_{uuid.uuid4().hex[:8]}"
        connected_client.create_graph(graph_name, GraphType.OPEN, "Test bulk import graph")
        yield graph_name
        drop_test_graph(connected_client, graph_name)

    def test_start_bulk_import(self, connected_client: GqldbClient, test_graph: str):
        """Test starting a bulk import session."""
        session = connected_client.start_bulk_import(test_graph, checkpoint_every=1000)
        assert session.success
        assert session.session_id != ""
        print(f"Started bulk import session: {session.session_id}")

        # Clean up - abort the session
        connected_client.abort_bulk_import(session.session_id)

    def test_bulk_import_workflow(self, connected_client: GqldbClient, test_graph: str):
        """Test the complete bulk import workflow."""
        # Step 1: Start bulk import session
        session = connected_client.start_bulk_import(test_graph, checkpoint_every=0)
        assert session.success
        print(f"Started bulk import session: {session.session_id}")

        # Step 2: Insert nodes with bulk import session
        nodes = [
            NodeData(labels=["BulkPerson"], properties={"name": "Alice", "idx": 1}),
            NodeData(labels=["BulkPerson"], properties={"name": "Bob", "idx": 2}),
            NodeData(labels=["BulkPerson"], properties={"name": "Charlie", "idx": 3}),
        ]

        node_result = connected_client.insert_nodes(
            test_graph,
            nodes,
            bulk_import_session_id=session.session_id
        )
        assert node_result.success
        assert node_result.node_count == 3
        print(f"Inserted {node_result.node_count} nodes")

        # Step 3: Get status
        status = connected_client.get_bulk_import_status(session.session_id)
        print(f"Bulk import status: active={status.is_active}, graph={status.graph_name}, records={status.record_count}")

        # Step 4: End bulk import
        end_result = connected_client.end_bulk_import(session.session_id)
        assert end_result.success
        print(f"Ended bulk import: {end_result.total_records} total records")

    def test_abort_bulk_import(self, connected_client: GqldbClient, test_graph: str):
        """Test aborting a bulk import session."""
        # Start bulk import session
        session = connected_client.start_bulk_import(test_graph, checkpoint_every=0)
        assert session.success
        print(f"Started bulk import session: {session.session_id}")

        # Insert some data
        nodes = [
            NodeData(labels=["AbortTest"], properties={"name": "TestNode"}),
        ]
        connected_client.insert_nodes(
            test_graph,
            nodes,
            bulk_import_session_id=session.session_id
        )

        # Abort the session
        abort_result = connected_client.abort_bulk_import(session.session_id)
        assert abort_result.success
        print("Aborted bulk import session successfully")

    def test_insert_edges_with_bulk_session(self, connected_client: GqldbClient, test_graph: str):
        """Test inserting edges with bulk import session."""
        # Start bulk import session
        session = connected_client.start_bulk_import(test_graph, checkpoint_every=0)
        assert session.success

        # Insert nodes
        nodes = [
            NodeData(labels=["Person"], properties={"name": "Alice"}),
            NodeData(labels=["Person"], properties={"name": "Bob"}),
        ]
        node_result = connected_client.insert_nodes(
            test_graph,
            nodes,
            bulk_import_session_id=session.session_id
        )
        assert node_result.success

        # Query to get node IDs (bulk import doesn't return IDs immediately)
        from gqldb.client import QueryConfig
        resp = connected_client.gql(
            "MATCH (n:Person) RETURN id(n) AS node_id ORDER BY n.name",
            QueryConfig(graph_name=test_graph)
        )

        node_ids = []
        for i in range(resp.row_count):
            row = resp.rows[i]
            value = resp.get_by_name(row, "node_id")
            if value:
                node_ids.append(value)

        if len(node_ids) >= 2:
            # Insert edges with bulk session
            edges = [
                EdgeData(
                    label="KNOWS",
                    from_node_id=node_ids[0],
                    to_node_id=node_ids[1],
                    properties={"since": 2024}
                ),
            ]
            edge_result = connected_client.insert_edges(
                test_graph,
                edges,
                bulk_import_session_id=session.session_id
            )
            assert edge_result.success
            print(f"Inserted {edge_result.edge_count} edges with bulk session")

        # End the session
        connected_client.end_bulk_import(session.session_id)



"""Connection management for GQLDB Python driver."""

from __future__ import annotations

import threading
import time
from dataclasses import dataclass
from typing import Dict, Optional

import grpc

from .config import GqldbConfig
from .errors import AllHostsFailedError, ConnectionClosedError, NoConnectionError


@dataclass
class Connection:
    """Represents a single gRPC connection."""
    host: str
    channel: grpc.Channel
    healthy: bool = True
    last_ping: float = 0.0


class ConnectionPool:
    """Manages a pool of connections to GQLDB servers."""

    def __init__(self, config: GqldbConfig):
        """Initialize the connection pool.

        Args:
            config: Configuration for connections.

        Raises:
            AllHostsFailedError: If all hosts fail to connect.
        """
        self._config = config
        self._connections: Dict[str, Connection] = {}
        self._lock = threading.RLock()
        self._closed = False
        self._stop_event = threading.Event()
        self._health_thread: Optional[threading.Thread] = None

        # Initialize connections to all hosts
        for host in config.hosts:
            try:
                conn = self._create_connection(host)
                self._connections[host] = conn
            except Exception:
                # Log error but continue with other hosts
                continue

        if not self._connections:
            raise AllHostsFailedError()

        # Start health check thread
        if config.health_check_interval > 0:
            self._health_thread = threading.Thread(
                target=self._health_check_loop,
                daemon=True
            )
            self._health_thread.start()

    def _create_connection(self, host: str) -> Connection:
        """Create a new gRPC connection to a host.

        Args:
            host: Host address in format "host:port".

        Returns:
            New Connection object.
        """
        options = [
            ("grpc.max_receive_message_length", self._config.max_recv_size),
            ("grpc.max_send_message_length", self._config.max_recv_size),
        ]

        if self._config.ssl_context:
            # Create secure channel with SSL
            credentials = grpc.ssl_channel_credentials()
            channel = grpc.secure_channel(host, credentials, options=options)
        else:
            # Create insecure channel
            channel = grpc.insecure_channel(host, options=options)

        return Connection(
            host=host,
            channel=channel,
            healthy=True,
            last_ping=time.time()
        )

    def get_connection(self) -> grpc.Channel:
        """Get a healthy connection from the pool.

        Returns:
            A gRPC channel.

        Raises:
            ConnectionClosedError: If the pool is closed.
            NoConnectionError: If no healthy connection is available.
        """
        with self._lock:
            if self._closed:
                raise ConnectionClosedError()

            # Find a healthy connection
            for conn in self._connections.values():
                if conn.healthy:
                    return conn.channel

            raise NoConnectionError()

    def get_connection_for_host(self, host: str) -> grpc.Channel:
        """Get a connection for a specific host.

        Args:
            host: Host address.

        Returns:
            A gRPC channel.

        Raises:
            NoConnectionError: If no connection is available for the host.
        """
        with self._lock:
            conn = self._connections.get(host)
            if not conn or not conn.healthy:
                raise NoConnectionError()
            return conn.channel

    def _health_check_loop(self) -> None:
        """Periodically check the health of all connections."""
        while not self._stop_event.wait(self._config.health_check_interval):
            self._check_health()

    def _check_health(self) -> None:
        """Check the health of all connections."""
        with self._lock:
            hosts = list(self._connections.keys())

        for host in hosts:
            with self._lock:
                conn = self._connections.get(host)
                if not conn:
                    continue

            try:
                # Check channel state
                state = conn.channel._channel.check_connectivity_state(False)
                healthy = state in (
                    grpc.ChannelConnectivity.READY,
                    grpc.ChannelConnectivity.IDLE
                )

                with self._lock:
                    conn.healthy = healthy
                    if healthy:
                        conn.last_ping = time.time()

                # Try to reconnect if unhealthy
                if not healthy:
                    self._reconnect(host)

            except Exception:
                with self._lock:
                    if conn:
                        conn.healthy = False

    def _reconnect(self, host: str) -> None:
        """Attempt to reconnect to a host.

        Args:
            host: Host address.
        """
        with self._lock:
            if self._closed:
                return

            # Close existing connection
            old_conn = self._connections.pop(host, None)
            if old_conn:
                try:
                    old_conn.channel.close()
                except Exception:
                    pass

            # Create new connection
            try:
                new_conn = self._create_connection(host)
                self._connections[host] = new_conn
            except Exception:
                pass

    def close(self) -> None:
        """Close all connections in the pool."""
        with self._lock:
            if self._closed:
                return

            self._closed = True
            self._stop_event.set()

            # Close all connections
            for conn in self._connections.values():
                try:
                    conn.channel.close()
                except Exception:
                    pass

            self._connections.clear()

        # Wait for health check thread
        if self._health_thread:
            self._health_thread.join(timeout=5.0)

    def healthy_host_count(self) -> int:
        """Return the number of healthy hosts.

        Returns:
            Number of healthy hosts.
        """
        with self._lock:
            return sum(1 for conn in self._connections.values() if conn.healthy)

    def hosts(self) -> list[str]:
        """Return the list of configured hosts.

        Returns:
            List of host addresses.
        """
        with self._lock:
            return list(self._connections.keys())

    @property
    def is_closed(self) -> bool:
        """Check if the pool is closed."""
        with self._lock:
            return self._closed



"""Transaction service handles transaction lifecycle management."""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable, List, TypeVar

import grpc

if TYPE_CHECKING:
    from ..transaction import Transaction
    from ..types import TransactionInfo
    from .service_context import ServiceContext

T = TypeVar('T')


class TransactionService:
    """Transaction service for managing database transactions."""

    def __init__(self, ctx: ServiceContext):
        self.ctx = ctx
        self._stub = None

    def _get_stub(self):
        if self._stub is None:
            from ..proto import gqldb_pb2_grpc
            channel = self.ctx.get_channel()
            self._stub = gqldb_pb2_grpc.TransactionServiceStub(channel)
        return self._stub

    def begin_transaction(self, graph_name: str, read_only: bool, timeout: int) -> Transaction:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        session = self.ctx.sessions.get_session()
        try:
            request = gqldb_pb2.BeginRequest(
                session_id=session.id if session else 0,
                graph_name=graph_name,
                read_only=read_only,
                timeout=timeout or self.ctx.config.timeout
            )
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().Begin(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            return self.ctx.tx_manager.begin(
                response.transaction_id,
                session.id if session else 0,
                graph_name,
                read_only,
                timeout or self.ctx.config.timeout
            )
        except grpc.RpcError as e:
            raise GqldbError(f"Begin transaction failed: {e.details()}") from e

    def commit(self, transaction_id: int) -> bool:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        session = self.ctx.sessions.get_session()
        try:
            request = gqldb_pb2.CommitRequest(
                session_id=session.id if session else 0,
                transaction_id=transaction_id
            )
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().Commit(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            self.ctx.tx_manager.commit(transaction_id)
            return response.success
        except grpc.RpcError as e:
            raise GqldbError(f"Commit failed: {e.details()}") from e

    def rollback(self, transaction_id: int) -> bool:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        session = self.ctx.sessions.get_session()
        try:
            request = gqldb_pb2.RollbackRequest(
                session_id=session.id if session else 0,
                transaction_id=transaction_id
            )
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().Rollback(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            self.ctx.tx_manager.rollback(transaction_id)
            return response.success
        except grpc.RpcError as e:
            raise GqldbError(f"Rollback failed: {e.details()}") from e

    def list_transactions(self) -> List[TransactionInfo]:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import TransactionInfo

        session = self.ctx.sessions.get_session()
        try:
            request = gqldb_pb2.ListTransactionsRequest(
                session_id=session.id if session else 0
            )
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().ListTransactions(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            return [
                TransactionInfo(
                    transaction_id=t.transaction_id,
                    session_id=t.session_id,
                    graph_name=t.graph_name,
                    read_only=t.read_only,
                    created_at=t.created_at,
                    duration_ms=t.duration_ms,
                    internal_tx_id=t.internal_tx_id
                )
                for t in response.transactions
            ]
        except grpc.RpcError as e:
            raise GqldbError(f"List transactions failed: {e.details()}") from e

    def with_transaction(
        self,
        graph_name: str,
        fn: Callable[[int], T],
        read_only: bool,
        timeout: int
    ) -> T:
        tx = self.begin_transaction(graph_name, read_only, timeout)
        try:
            result = fn(tx.id)
            self.commit(tx.id)
            return result
        except Exception as e:
            self.rollback(tx.id)
            raise e
